"""
Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

import datetime
import sys
from urllib.parse import urljoin

# Sphinx extention to format xarray/pandas summaries
import sphinx_autosummary_accessors
from jinja2.defaults import DEFAULT_FILTERS
from packaging.version import Version

import imas_paraview

print("python exec:", sys.executable)
print("sys.path:", sys.path)

# -- Project information -----------------------------------------------------
# The documented project’s name
project = src_project = PROJECT = "IMAS-ParaView"
PACKAGE = "imas_paraview"

# A copyright statement in the style '2008, Author Name'.
copyright = f"2022-{datetime.datetime.now().year}, ITER Organization"
# The author name(s) of the document
author = "ITER Organization"
src_host = "https://github.com/iterorganization/"

# Parse urls here for convenience, to be re-used

# ITER docs
iter_projects = "https://github.com/iterorganization/"
dd_url = urljoin(iter_projects, "IMAS-Data-Dictionary/")
al_url = urljoin(iter_projects, "IMAS-Python/")
issue_url = jira_url = "https://github.com/iterorganization/IMAS-ParaView/issues"

# IMAS-ParaView
repository_url = f"{iter_projects}/{src_project}/"
blob_url = repository_url
mr_url = urljoin(repository_url, "/pulls")


# Configuration of sphinx.ext.extlinks
# See https://www.sphinx-doc.org/en/master/usage/extensions/extlinks.html
# unique name: (base URL, label prefix)
extlinks = {
    "src": (blob_url + "%s", "%s"),
    "issue": (issue_url + "%s", "%s"),
    "merge": (mr_url + "%s", "!%s"),
    "dd": (dd_url + "%s", "%s"),
    "al": (al_url + "%s", "%s"),
    "pypa": ("https://packaging.python.org/%s", None),

}

full_version = Version(imas_paraview.__version__)

# version: The major project version, used as the replacement for |version|.
#   For example, for the Python documentation, this may be something like 2.6.
version = full_version.base_version

# release: The full project version, used as the replacement for |release| and
#   e.g. in the HTML templates. For example, for the Python documentation, this
#   may be something like 2.6.0rc1
release = str(full_version)


# -- General configuration ---------------------------------------------------
# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",  # To auto-generate docs from Python docstrings
    "sphinx.ext.todo",  # Support for todo items
    "sphinx.ext.napoleon",  # Support for NumPy and Google style docstrings
    "sphinx.ext.intersphinx",  # Generate links to other documentation files
    "sphinx.ext.autosummary",  # For summarizing autodoc-generated files
    "sphinx.ext.extlinks",  # For shortening internal links
    "sphinx.ext.mathjax",  # Render math as images
    "sphinx_immaterial",  # Sphinx immaterial theme
]

todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates", sphinx_autosummary_accessors.templates_path]

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = "%Y-%m-%d"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_immaterial"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
# and
# https://sphinx-immaterial.readthedocs.io/en/latest/customization.html#confval-html_theme_options
html_theme_options = {
    "repo_url": "https://github.com/iterorganization/IMAS-Paraview",
    "repo_name": "IMAS-ParaView",
    "icon": {
        "repo": "fontawesome/brands/github",
    },
    "features": [
        # "navigation.expand",
        # "navigation.tabs",
        "navigation.sections",
        "navigation.instant",
        # "header.autohide",
        "navigation.top",
        # "navigation.tracking",
        # "search.highlight",
        # "search.share",
        # "toc.integrate",
        "toc.follow",
        "toc.sticky",
        # "content.tabs.link",
        "announce.dismiss",
    ],
    # "toc_title_is_page_title": True,
    # "globaltoc_collapse": True,
    "palette": [
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "primary": "indigo",
            "accent": "green",
            "toggle": {
                "icon": "material/lightbulb-outline",
                "name": "Switch to dark mode",
            },
        },
        {
            "media": "(prefers-color-scheme: dark)",
            "scheme": "slate",
            "primary": "light-blue",
            "accent": "lime",
            "toggle": {
                "icon": "material/lightbulb",
                "name": "Switch to light mode",
            },
        },
    ],
}

# Shorten Table Of Contents in API documentation
object_description_options = [
    (".*", dict(include_fields_in_toc=False)),
    (".*parameter", dict(include_in_toc=False)),
]

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# '<project> v<release> documentation'.
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = "_static/imaspy_200x200.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named 'default.css' will overwrite the builtin 'default.css'.
html_static_path = ["_static"]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = today_fmt

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, 'Created using Sphinx' is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, '(C) Copyright ...' is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. '.xhtml').
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "ggd_vtk_doc"


# -- Extension configuration -------------------------------------------------
# Configuration of sphinx.ext.autodoc
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html
autodoc_typehints = "signature"

# The documentation for these external import give errors, so we mock them
autodoc_mock_imports = ["vtkmodules", "paraview", "vtk"]

# Configuration of sphinx.ext.autosummary
# https://www.sphinx-doc.org/en/master/usage/extensions/autosummary.html
autosummary_generate = True


# Configuration of sphinx.ext.napoleon
# https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html
# Support for NumPy and Google style docstrings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
# Allow PEP 526 attributes annotations in classes:
napoleon_attr_annotations = True


# Configuration of sphinx.ext.intersphinx
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "packaging": ("https://packaging.pypa.io/en/stable/", None),
}
intersphinx_timeout = 60  # Downloads time out after 1 minute

smartquotes = False

# Configuration of sphinx.ext.mathjax
# https://www.sphinx-doc.org/en/master/usage/extensions/math.html#module-sphinx.ext.mathjax


def escape_underscores(string):
    return string.replace("_", r"\_")


def setup(app):
    DEFAULT_FILTERS["escape_underscores"] = escape_underscores
    app.add_css_file("imas_paraview.css")



"""Helper functions for generating IMAS URIs."""

from pathlib import Path
from typing import Optional

import imas

BACKENDS = {
    imas.ids_defs.MDSPLUS_BACKEND: "mdsplus",
    imas.ids_defs.HDF5_BACKEND: "hdf5",
    imas.ids_defs.ASCII_BACKEND: "ascii",
}
"""Mapping of backend ID to the backend string used in the URI."""


def uri_from_path(path: Optional[str]) -> Optional[str]:
    """Create an IMAS URI from a local path.

    Args:
        path: Path to a local file. This should point to a valid file from the HDF5
            backend (with extension ``.h5``) or MDSplus backend (with extension
            ``.characteristics``, ``.datafile`` or ``.tree``).

    Returns:
        An IMAS URI pointing to the local file, or None if no URI could be constructed
        from the provided file.
    """
    uri = None
    if path:
        path = Path(path)
        if path.suffix == ".h5":
            uri = f"imas:hdf5?path={path.parent}"
        elif path.suffix in [".characteristics", ".datafile", ".tree"]:
            uri = f"imas:mdsplus?path={path.parent}"
        elif path.suffix == ".ids":
            uri = f"imas:ascii?path={path.parent}"
    return uri


def uri_from_pulse_run(
    backend: int, database: str, pulse: int, run: int, user: str, version: str
) -> Optional[str]:
    """Create an IMAS URI from 'legacy' parameters.

    Args:
        backend: Backend ID, e.g. HDF5_BACKEND
        database: Database name, e.g. "ITER"
        pulse: Pulse number of the database entry
        run: Run number of the database entry
        user: User name of the database
        version: Major version of the DD used by the the access layer

    Returns:
        An IMAS URI pointing to the data entry, or None if no valid URI could be
        constructed.
    """
    if backend not in BACKENDS:
        return None
    if not database or not user or not version:
        return None
    # TODO: use imas_core._al_lowlevel.al_build_uri_from_legacy_parameters instead?
    return (
        f"imas:{BACKENDS[backend]}?user={user};pulse={pulse};run={run};"
        f"database={database};version={version}"
    )



"""Plugin to view line of sight IDS structures in Paraview."""

import logging
from dataclasses import dataclass

from imas.ids_structure import IDSStructure
from paraview.util.vtkAlgorithm import smhint, smproxy
from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkCommonDataModel import (
    vtkCellArray,
    vtkLine,
    vtkMultiBlockDataSet,
    vtkPolyData,
)

from imas_paraview.ids_util import get_object_by_name
from imas_paraview.paraview_support.servermanager_tools import (
    doublevector,
    propertygroup,
)
from imas_paraview.plugins.base_class import GGDVTKPluginBase
from imas_paraview.util import pol_to_cart

logger = logging.getLogger("imas_paraview")

SUPPORTED_LINE_OF_SIGHT_IDS = [
    "bolometer",
    "bremsstrahlung_visible",
    "ece",
    "hard_x_rays",
    "interferometer",
    "mse",
    "polarimeter",
    "refractometer",
    "soft_x_rays",
    "spectrometer_uv",
    "spectrometer_visible",
]


@dataclass
class LineOfSight:
    """Data class that stores ``line_of_sight`` IDS structures, along with its name."""

    name: str
    line_of_sight: IDSStructure


@smproxy.source(label="Line of Sight Reader")
@smhint.xml("""<ShowInMenu category="IMAS Tools" />""")
class LineOfSightReader(GGDVTKPluginBase):
    """Line of sight reader based on IMAS-Python"""

    def __init__(self):
        super().__init__("vtkMultiBlockDataSet", SUPPORTED_LINE_OF_SIGHT_IDS)
        self.scaling_factor = 1

    @doublevector(label="Scaling Factor", name="scaling_factor", default_values=1)
    def P99_SetScalingFactor(self, val):
        """Sets the scaling factor for the line of sight's length."""
        self._update_property("scaling_factor", val)

    @propertygroup("Line of Sight Settings", ["scaling_factor"])
    def PG3_LineOfSightGroup(self):
        """Dummy function to define a PropertyGroup."""

    def GetAttributeArrayName(self, idx) -> str:
        return self._selectable[idx].name

    def RequestData(self, request, inInfo, outInfo):
        if self._dbentry is None or not self._ids_and_occurrence or self._ids is None:
            return 1

        if len(self._selected) > 0:
            output = vtkMultiBlockDataSet.GetData(outInfo)
            self._load_los(output)
        return 1

    def setup_ids(self):
        """
        Select which line_of_sights to show in the array domain selector. The
        channel AoS is search through for line_of_sight structures. Their names are
        generated based on the channel name, and added to the array domain selector.
        """
        assert self._ids is not None, "IDS cannot be empty during setup."

        self._selectable = []

        channels = self._ids.channel

        for i, channel in enumerate(channels):
            channel_name = channel.name
            if channel_name == "":
                channel_name = f"channel {i}"
                logger.warning(
                    "Found a channel without a name, "
                    f"it will be loaded as {channel_name}."
                )

            # For ece, if the channel does not have a line of sight, or it is empty,
            # use the "global" line_of_sight instead
            if self._ids.metadata.name == "ece" and (
                not hasattr(channel, "line_of_sight")
                or not channel.line_of_sight.first_point.r.has_value
            ):
                selectable = LineOfSight(str(channel_name), self._ids.line_of_sight)
            else:
                selectable = LineOfSight(str(channel_name), channel.line_of_sight)
            self._selectable.append(selectable)

    def _load_los(self, output):
        """Go through the list of selected line_of_sights, and load each of them in a
        separate vtkPolyData object, which are all combined into a vtkMultiBlockDataSet.

        Args:
            output: The vtkMultiBlockDataSet containing the line_of_sights.
        """
        for i, channel_name in enumerate(self._selected):
            channel = get_object_by_name(self._selectable, channel_name)
            if channel is None:
                raise ValueError(f"Could not find {channel_name}")

            logger.info(f"Selected {channel.name}")
            vtk_poly = self._create_vtk_los(channel)
            output.SetBlock(i, vtk_poly)

    def _extend_line(self, first_point, second_point):
        """Extends a line in the direction from the first_point to the second_point. The
        distance from the first_point to the new point is determined by the original
        distance between the first and second point, scaled by the scaling factor.

        Args:
            first_point: Tuple containing the (x,y,z) coordinates of the first point.
            second_point: Tuple containing the (x,y,z) coordinates of the second point.

        Returns:
            Tuple containing a point (x,y,z) along the line from the first to the second
            point, at a distance from the first point defined by the scaling factor.
        """
        line1_direction = tuple(x - y for x, y in zip(second_point, first_point))
        return tuple(
            x + self.scaling_factor * y for x, y in zip(first_point, line1_direction)
        )

    def _create_vtk_los(self, channel):
        """Create a vtkPolyData containing a line, based on the r, phi, and z
        coordinates in the line of sight structures. The r-, phi-, and z-coordinates are
        converted to cartesian and stored as vtkPoints and connected using vtkLines,
        which are both stored in a vtkPolyData object. The lines can be scaled by a
        scaling factor.

        The connecting lines are created based on the graph below. Here, the points
        0 -> 1 describe the main LoS (los.first_point -> los.second_point). The
        reflection is described by points 1 -> 2 (los.second_point -> los.third_point).
        The extensions are visualized by the plusses in the graph, from 0 -> 3 for the
        main LoS, and from 1 -> 4, for the reflection. Only the lines from 0 -> 3 (and
        1 -> 4, if there is a reflection) are transformed into vtkLines.

        0----1+++++3
              \
               \
                2
                 +
                  +
                   4
        Args:
            channel: containing a line_of_sight structure.

        Returns:
            vtkPolyData containing line_of_sight data.
        """
        los = channel.line_of_sight
        points = [None] * 5
        points[0] = los.first_point
        points[1] = los.second_point

        # Add reflection point
        has_reflection = False
        if hasattr(los, "third_point"):
            third_point = los.third_point
            if (
                third_point.r.has_value
                and third_point.phi.has_value
                and third_point.z.has_value
            ):
                has_reflection = True
                points[2] = third_point

        # Convert from cylindrical to cartesian coordinates
        points = [
            (*pol_to_cart(point.r, point.phi), point.z) if point is not None else None
            for point in points
        ]

        # Add line extension
        points[3] = self._extend_line(points[0], points[1])
        if has_reflection:
            points[4] = self._extend_line(points[1], points[2])

        vtk_poly = self._create_vtk_poly(points, has_reflection)
        return vtk_poly

    def _create_vtk_poly(self, points, has_reflection):
        """Create vtkPolyData containing the vtkPoints and vtkLines in a vtkCellArray.

        Args:
            points: List of points to store in the vtkPolyData.
            has_reflection: boolean whether LoS contains a reflection point.

        Returns:
            vtkPolyData containing vtkPoints and vtkCellArray containing the vtkLines
            corresponding to the scaled LoS and optionally its reflection.
        """
        vtk_points = vtkPoints()
        vtk_lines = vtkCellArray()

        vtk_points.InsertNextPoint(*points[0])
        vtk_points.InsertNextPoint(*points[3])
        line = vtkLine()
        line.GetPointIds().SetId(0, 0)
        line.GetPointIds().SetId(1, 1)
        vtk_lines.InsertNextCell(line)
        if has_reflection:
            vtk_points.InsertNextPoint(*points[1])
            vtk_points.InsertNextPoint(*points[4])
            line = vtkLine()
            line.GetPointIds().SetId(0, 2)
            line.GetPointIds().SetId(1, 3)
            vtk_lines.InsertNextCell(line)

        vtk_poly = vtkPolyData()
        vtk_poly.SetPoints(vtk_points)
        vtk_poly.SetLines(vtk_lines)
        return vtk_poly



# SPDX-FileCopyrightText: Copyright (c) Kitware Inc.
# SPDX-License-Identifier: BSD-3-Clause

import logging

from vtkmodules.vtkCommonCore import vtkLogger
from vtkmodules.vtkCommonCore import vtkOutputWindow as win


class VTKHandler(logging.Handler):
    """
    A custom logging handler for VTK, intended to work independently from ParaView,
    adapted from paraview.detail.loghandler.py.

    This handler formats and logs messages using VTK's logging system and displays
    them in the VTK output window. It maps Python's standard logging levels to VTK's
    verbosity levels.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def emit(self, record):
        """Formats and logs a record using VTK's logging system and displays it in the
        VTK output window.

        Args:
            record: The log record to be processed.

        Raises:
            Exception: Any exceptions raised during the logging process will be
            handled by the `handleError` method.
        """
        try:
            msg = self.format(record)
            lvl = self.get_vtk_level(record.levelno)
            vtkLogger.Log(lvl, record.filename, record.lineno, msg)

            outputWindow = win.GetInstance()
            if outputWindow:
                # do not duplicate on standard output
                prevMode = outputWindow.GetDisplayMode()
                outputWindow.SetDisplayModeToNever()

                if lvl == vtkLogger.VERBOSITY_ERROR:
                    lvlText = "ERR: "
                    fullMsg = f"{record.filename}:{record.lineno} {lvlText}{msg}\n"
                    outputWindow.DisplayErrorText(fullMsg)
                elif lvl == vtkLogger.VERBOSITY_WARNING:
                    lvlText = "WARN: "
                    fullMsg = f"{record.filename}:{record.lineno} {lvlText}{msg}\n"
                    outputWindow.DisplayWarningText(fullMsg)
                elif lvl == vtkLogger.VERBOSITY_INFO:
                    lvlText = "INFO: "
                    fullMsg = f"{record.filename}:{record.lineno} {lvlText}{msg}\n"
                    outputWindow.DisplayText(fullMsg)
                elif lvl == vtkLogger.VERBOSITY_TRACE:
                    lvlText = "DEBUG: "
                    fullMsg = f"{record.filename}:{record.lineno} {lvlText}{msg}\n"
                    outputWindow.DisplayDebugText(fullMsg)
                else:
                    fullMsg = f"{record.filename}:{record.lineno} {msg}\n"
                    outputWindow.DisplayText(fullMsg)

                outputWindow.SetDisplayMode(prevMode)

        except Exception:
            self.handleError(record)

    def get_vtk_level(self, level):
        """Converts a Python logging level to a VTK verbosity level.

        Args:
            level: The Python logging level (e.g., logging.ERROR, logging.WARNING).

        Returns:
            The corresponding VTK verbosity level.
        """
        if level >= logging.ERROR:
            return vtkLogger.VERBOSITY_ERROR
        elif level >= logging.WARNING:
            return vtkLogger.VERBOSITY_WARNING
        elif level >= logging.INFO:
            return vtkLogger.VERBOSITY_INFO
        elif level >= logging.DEBUG:
            return vtkLogger.VERBOSITY_TRACE
        else:
            return vtkLogger.VERBOSITY_MAX

    def get_level(self, vtklevel=None):
        """Converts a VTK verbosity level to a Python logging level. If VTK verbosity
        level is provided, it defaults to using the current verbosity cutoff defined in
        the VTK logger.

        Args:
            vtklevel: The VTK verbosity level to convert.

        Returns:
            The corresponding Python logging level.
        """
        from vtkmodules.vtkCommonCore import vtkLogger

        if vtklevel is None:
            vtk_level = vtkLogger.GetCurrentVerbosityCutoff()

        if vtk_level == vtkLogger.VERBOSITY_ERROR:
            return logging.ERROR
        elif vtk_level == vtkLogger.VERBOSITY_WARNING:
            return logging.WARNING
        elif vtk_level == vtkLogger.VERBOSITY_INFO:
            return logging.INFO
        else:
            return logging.DEBUG



"""
A collection of methods to read a grid_ggd IDS node into a VTK dataset.
These methods copy contents from the grid_ggd/space and grid_ggd/grid_subset
children into distinct vtkUnstructuredGrid objects for Bezier elements.
"""

import logging

import numpy as np
import vtk
from vtkmodules.util import numpy_support as npvtk
from vtkmodules.util.vtkConstants import VTK_DOUBLE
from vtkmodules.vtkCommonDataModel import vtkUnstructuredGrid

logger = logging.getLogger("imas_paraview")

prec = np.float64
vtk_prec = VTK_DOUBLE


def read_plasma_state(grid_ggd, ps_reader, plane_config, ugrid):
    """Reads plasma state data arrays from the ggd node in the IDS. These arrays are
        added as point data to the unstructured grid.

    Args:
        grid_ggd: The GGD grid.
        ps_reader: PlasmaStateReader which contains which array to load
        plane_config: Configuration for the bezier planes
        ugrid: The ugrid to load the GGD arrays for
    """
    n_plane = plane_config.n_plane

    phi = [plane_config.phi_start, plane_config.phi_end]
    val_tor1 = np.array([])
    nam = list()

    N_vertex = len(grid_ggd.space[0].objects_per_dimension[0].object)
    array_list = ps_reader.scalar_array_list + ps_reader.vector_array_list

    # Load selected arrays
    for attribute_array in array_list:
        name = ps_reader._create_name_with_units(attribute_array)
        if hasattr(attribute_array[0], "coefficients"):
            scalar_data = attribute_array[0].coefficients
        elif hasattr(attribute_array[0], "phi_coefficients"):
            scalar_data = attribute_array[0].phi_coefficients
        else:
            logger.warning(f"Could not load the coefficients for {name}, it is ignored")
            continue
        nam.append(name)
        if np.size(val_tor1) == 0:
            val_tor1 = np.array([scalar_data])
        else:
            val_tor1 = np.concatenate((val_tor1, np.array([scalar_data])), axis=0)

    n_val = len(nam)
    a = np.shape(val_tor1)
    n_tor = a[1] // N_vertex if len(a) > 1 else 0

    valu = np.reshape(val_tor1, (n_val, n_tor, N_vertex, 4))
    values = np.swapaxes(valu, 2, 3)
    values = np.swapaxes(values, 1, 2)

    n_period = grid_ggd.space[1].geometry_type.index
    n_sub = 4
    without_n0_mode = False

    n_plane = 1 + (n_plane - 1) * 2
    phis = calc_phis(n_plane, phi)

    HZ = toroidal_basis(n_tor, n_period, phis, without_n0_mode)

    vertex = get_vertex(grid_ggd)
    size = calc_size(grid_ggd)
    val = interp_scalars_3D(values, vertex, size, n_sub, HZ)
    if n_val > 0:
        val = val.reshape((n_val, -1))
    a = np.shape(val)
    val = val.reshape((a[0], a[1] // 16, 16))
    val[:, :, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]] = val[
        :, :, [0, 12, 15, 3, 4, 8, 11, 7, 1, 13, 14, 2, 5, 9, 10, 6]
    ]
    val = val.reshape((a[0], a[1]))

    for i in range(len(nam)):
        tmp = npvtk.numpy_to_vtk(val[i, :], deep=True, array_type=vtk_prec)
        tmp.SetName(nam[i])
        ugrid.GetPointData().AddArray(tmp)


def convert_grid_subset_to_unstructured_grid(
    grid_ggd, plane_config
) -> vtkUnstructuredGrid:
    """
    Copy the elements found in given grid_ggd/grid_subset IDS node into a
    vtkUnstructuredGrid instance. This method uses the supplied point coordinates in
    the form of a vtkPoints instance.

    Args:
        grid_ggd: The GGD grid.
        plane_config: Configuration for the bezier planes.

    Returns:
        The created unstructured grid.
    """
    output = vtkUnstructuredGrid()

    n_plane = plane_config.n_plane
    phi = [plane_config.phi_start, plane_config.phi_end]
    gr2d = grid_ggd.space[0]
    N_vertex = len(gr2d.objects_per_dimension[0].object)

    x = np.zeros((2, 4, N_vertex))
    for j, obj in enumerate(gr2d.objects_per_dimension[0].object):
        x[:, :, j] = obj.geometry_2d.value

    size = calc_size(grid_ggd)
    vertex = get_vertex(grid_ggd)

    # Everything we need to visualise data is now excavated from IDS file
    n_plane = 1 + (n_plane - 1) * 2

    phis = calc_phis(n_plane, phi)
    ien = None

    tmp = np.zeros((x.shape[0], x.shape[1], vertex.shape[0], vertex.shape[1]))
    for i in range(vertex.shape[0]):  # small loop over vertices (hardcode 4 here?)
        tmp[:, :, i, :] = x[:, :, vertex[i, :] - 1]
    # multiply by size[order, vertex, element]
    tmp[0, :, :, :] *= size
    tmp[1, :, :, :] *= size
    # Create output array
    xy = np.zeros((np.shape(tmp)[3] * 16, 2))
    for i in range(np.shape(tmp)[3]):
        x1 = tmp[0, :, :, i]
        y1 = tmp[1, :, :, i]
        x1[3, :] = x1[3, :] + x1[1, :] + x1[2, :]
        y1[3, :] = y1[3, :] + y1[1, :] + y1[2, :]
        x1[1:, :] += np.tile(x1[0, :], (3, 1))
        y1[1:, :] += np.tile(y1[0, :], (3, 1))
        xy[16 * i : 16 * (i + 1), 0] = np.ravel(x1)
        xy[16 * i : 16 * (i + 1), 1] = np.ravel(y1)
    RZ = xy

    n_xy = np.shape(RZ)[0]
    xyz = np.zeros((n_xy * n_plane, 3))
    for i in range(n_plane):
        # x = R * cos(phi); y = R * sin(phi), z = Z
        xyz[i * n_xy : (i + 1) * n_xy, 0] = np.ravel(RZ[:, 0] * np.cos(phis[i]))
        xyz[i * n_xy : (i + 1) * n_xy, 1] = np.ravel(RZ[:, 0] * np.sin(phis[i]))
        xyz[i * n_xy : (i + 1) * n_xy, 2] = np.ravel(RZ[:, 1])

    if n_plane == 1:
        index = np.array([0, 1, 2, 3, 4, 5, 9, 10, 7, 6, 8, 11, 12, 13, 15, 14])
        step = np.array([16 for i in range(16)])
        ien2 = np.array([index])
        for i in range(1, np.shape(xyz)[0] // 16):
            ien2 = np.concatenate((ien2, np.array([index + i * step])), axis=0)
        ien = np.insert(ien2, 0, 16, axis=1)
        etype = vtk.VTK_BEZIER_QUADRILATERAL

    else:
        alpha = (phi[1] - phi[0]) / (n_plane - 1)
        w = np.cos(np.deg2rad(alpha))
        w1 = np.ones((np.shape(xyz)[0]))
        ien = None
        # Connectivity list of one bezier cell:
        index = np.array(
            [
                0,
                1,
                2,
                3,
                0 + 2 * n_xy,
                1 + 2 * n_xy,
                2 + 2 * n_xy,
                3 + 2 * n_xy,
                4,
                5,
                9,
                10,
                7,
                6,
                8,
                11,
                4 + 2 * n_xy,
                5 + 2 * n_xy,
                9 + 2 * n_xy,
                10 + 2 * n_xy,
                7 + 2 * n_xy,
                6 + 2 * n_xy,
                8 + 2 * n_xy,
                11 + 2 * n_xy,
                # According to VTK docs, the sequence should be ([0, 1, 3, 2] + n_xy).
                # However that doesn't render correctly and we need to swap the 3 and
                # the 2... See also
                # https://www.kitware.com/main/wp-content/uploads/2020/03/Implementation-of-rational-Be%CC%81zier-cells-into-VTK-Report.pdf
                0 + n_xy,
                1 + n_xy,
                2 + n_xy,
                3 + n_xy,
                8 + n_xy,
                11 + n_xy,
                9 + n_xy,
                10 + n_xy,
                4 + n_xy,
                5 + n_xy,
                7 + n_xy,
                6 + n_xy,
                12,
                13,
                15,
                14,
                12 + 2 * n_xy,
                13 + 2 * n_xy,
                15 + 2 * n_xy,
                14 + 2 * n_xy,
                12 + n_xy,
                13 + n_xy,
                15 + n_xy,
                14 + n_xy,
            ]
        )
        # Connectivity for all cells between two planes
        repeat_in_plane = 16 * np.arange(n_xy // 16)
        repeat_planes = np.arange((n_plane - 1) // 2) * 2 * n_xy
        repeat_index = (repeat_in_plane + repeat_planes[:, np.newaxis]).ravel()
        ien = index + repeat_index[:, np.newaxis]

        # set rational weights of midplane points and adjust x/y values accordingly
        for i in range((n_plane - 1) // 2):
            w1[n_xy * (1 + 2 * i) : n_xy * (2 + 2 * i)] = w
            xyz[n_xy * (1 + 2 * i) : n_xy * (2 + 2 * i), 0:2] /= w

        # VTK cell list requires number of points as first element:
        ien = np.insert(ien, 0, 48, axis=1)
        etype = vtk.VTK_BEZIER_HEXAHEDRON

        weights = npvtk.numpy_to_vtk(w1, deep=True, array_type=vtk_prec)
        weights.SetName("RationalWeights")
        output.GetPointData().SetRationalWeights(weights)

        # Bezier degrees of the cells
        degrees = npvtk.numpy_to_vtk(
            np.repeat([[3, 3, 2]], len(ien), axis=0),
            deep=True,
            array_type=vtk.VTK_ID_TYPE,
        )
        degrees.SetName("HigherOrderDegrees")
        output.GetCellData().SetHigherOrderDegrees(degrees)

    pcoords = npvtk.numpy_to_vtk(xyz, deep=True, array_type=vtk_prec)
    points = vtk.vtkPoints()
    points.SetData(pcoords)

    cells = vtk.vtkCellArray()
    cells.SetCells(
        ien.shape[0], npvtk.numpy_to_vtk(ien, deep=True, array_type=vtk.VTK_ID_TYPE)
    )

    output.SetPoints(points)
    output.SetCells(etype, cells)

    return output


def calc_size(grid_ggd):
    """Calculate the size matrix for each face in the grid.
    Uses the geometry values from the 2D grid representation.

    Args:
        grid_ggd: The GGD grid.

    Returns:
        The calculated size matrix.
    """
    # size 1, d_{uk}, d_{vk}, d{uv}d{vk} as in Daan Van Vugt thesis
    gr2d = grid_ggd.space[0]
    n_face = len(gr2d.objects_per_dimension[2].object)
    size = np.empty((4, 4, n_face))

    for i, obj in enumerate(gr2d.objects_per_dimension[2].object):
        size[:, :, i] = obj.geometry_2d.value
    return size


def get_vertex(grid_ggd):
    """Retrieve vertex numbers in the R-Z plane.
    Extracts node indices for each face in the grid.

    Args:
        grid_ggd: The GGD grid.

    Returns:
        The vertex numbers.
    """
    # Get vertex numbers in the R,Z plane
    object_aos = grid_ggd.space[0].objects_per_dimension[2].object
    vertex = np.zeros((len(object_aos[0].nodes), len(object_aos)), dtype=np.int32)
    for i, obj in enumerate(object_aos):
        vertex[:, i] = obj.nodes.value
    return vertex


def calc_phis(n_plane, phi):
    """Compute an array of phi values for given planes.
    Determines if the values are periodic and converts them to radians.

    Args:
        n_plane: The number of planes.
        phi: Array containing the start and end phi values.

    Returns:
        the array of phi values for the planes.
    """
    periodic = False
    if n_plane == 1:
        phis = np.asarray([phi[0]])
    else:
        periodic = np.mod(phi[0] - phi[1], 360) < 1e-9
        phis = np.linspace(phi[0], phi[1], num=n_plane, endpoint=not periodic)

    phis = phis * np.pi / 180
    return phis


def toroidal_basis(n_tor, n_period, phis, without_n0_mode):
    # Setup toroidal coefficients for each plane and toroidal harmonic
    HZ = np.zeros((n_tor, len(phis)))
    for i in range(n_tor):
        mode = np.floor((i + 1) / 2) * n_period
        if i == 0:
            if not without_n0_mode:
                HZ[i, :] = 1
        elif i % 2 == 0:
            HZ[i, :] = np.sin(mode * phis)
        elif i % 2 == 1:
            HZ[i, :] = np.cos(mode * phis)
    return HZ


def interp_scalars(values, vertex, size, n_sub):
    """
    Interpolate scalars on 2D poloidal plane

    returns:
        values: interpolated values, values[var, harmonic, element, is, it]
    """
    # Multiply values[var,order,harm,vertex,element] with
    # size[order, vertex, element] and bf[order, vertex, s, t]
    return np.einsum(
        "lihjk,ijk,ijmn->lhkmn", values[:, :, :, vertex - 1], size, bf(n_sub)
    )


def interp_scalars_3D(values, vertex, size, n_sub, HZ):
    """
    Interpolate scalars on 2D planes * n_planes
    """
    vals = interp_scalars(values, vertex, size, n_sub)
    return np.einsum("lhkmn,hp->lpkmn", vals, HZ)


def basis_functions(s, t):
    """
    Calculate values of the basis functions at positions s and t
    Optionally put many values of s and t at once as numpy arrays.
    Dimension 0: order
    Dimension 1: vertex
    optional dimension 2, 3: position s, t
    """
    return np.asarray(
        [
            [
                (-1 + s) ** 2 * (1 + 2 * s) * (-1 + t) ** 2 * (1 + 2 * t),
                -(s**2 * (-3 + 2 * s) * (-1 + t) ** 2 * (1 + 2 * t)),
                s**2 * (-3 + 2 * s) * t**2 * (-3 + 2 * t),
                -((-1 + s) ** 2) * (1 + 2 * s) * t**2 * (-3 + 2 * t),
            ],
            [
                3 * (-1 + s) ** 2 * s * (-1 + t) ** 2 * (1 + 2 * t),
                -3 * (-1 + s) * s**2 * (-1 + t) ** 2 * (1 + 2 * t),
                3 * (-1 + s) * s**2 * t**2 * (-3 + 2 * t),
                -3 * (-1 + s) ** 2 * s * t**2 * (-3 + 2 * t),
            ],
            [
                3 * (-1 + s) ** 2 * (1 + 2 * s) * (-1 + t) ** 2 * t,
                -3 * s**2 * (-3 + 2 * s) * (-1 + t) ** 2 * t,
                3 * s**2 * (-3 + 2 * s) * (-1 + t) * t**2,
                -3 * (-1 + s) ** 2 * (1 + 2 * s) * (-1 + t) * t**2,
            ],
            [
                9 * (-1 + s) ** 2 * s * (-1 + t) ** 2 * t,
                -9 * (-1 + s) * s**2 * (-1 + t) ** 2 * t,
                9 * (-1 + s) * s**2 * (-1 + t) * t**2,
                -9 * (-1 + s) ** 2 * s * (-1 + t) * t**2,
            ],
        ]
    )


def basis_functions_s(s, t):
    """
    Calculate values of the basis functions derived to s at positions s and t
    Optionally put many values of s and t at once as numpy arrays.
    Dimension 0: order
    Dimension 1: vertex
    optional dimension 2, 3: position s, t
    """
    return np.asarray(
        [
            [
                6 * (-1 + s) * s * (-1 + t) ** 2 * (1 + 2 * t),
                -6 * (-1 + s) * s * (-1 + t) ** 2 * (1 + 2 * t),
                6 * (-1 + s) * s * t**2 * (-3 + 2 * t),
                -6 * (-1 + s) ** 2 * t**2 * (-3 + 2 * t),
            ],
            [
                3 * (-1 + s) * (-1 + 3 * s) * (-1 + t) ** 2 * (1 + 2 * t),
                -3 * s * (-2 + 3 * s) * (-1 + t) ** 2 * (1 + 2 * t),
                3 * s * (-2 + 3 * s) * t**2 * (-3 + 2 * t),
                -3 * (-1 + 3 * s) * (-1 + s) * t**2 * (-3 + 2 * t),
            ],
            [
                18 * (-1 + s) * s * (-1 + t) ** 2 * t,
                -18 * (-1 + s) * s * (-1 + t) ** 2 * t,
                18 * (-1 + s) * s * (-1 + t) * t**2,
                -18 * (-1 + s) * s * (-1 + t) * t**2,
            ],
            [
                9 * (-1 + s) * (-1 + 3 * s) * (-1 + t) ** 2 * t,
                -9 * s * (-2 + 3 * s) * (-1 + t) ** 2 * t,
                9 * s * (-2 + 3 * s) * (-1 + t) * t**2,
                -9 * (-1 + 3 * s) * (-1 + s) * (-1 + t) * t**2,
            ],
        ]
    )


def basis_functions_t(s, t):
    """
    Calculate values of the basis functions derived to t at positions s and t
    Optionally put many values of s and t at once as numpy arrays.
    Dimension 0: order
    Dimension 1: vertex
    optional dimension 2, 3: position s, t
    """
    return np.asarray(
        [
            [
                6 * (-1 + s) ** 2 * (1 + 2 * s) * (-1 + t) * t,
                -6 * s**2 * (-3 + 2 * s) * (-1 + t) * t,
                6 * s**2 * (-3 + 2 * s) * (-1 + t) * t,
                -6 * (-1 + s) ** 2 * (1 + 2 * s) * (-1 + t) * t,
            ],
            [
                18 * (-1 + s) ** 2 * s * (-1 + t) * t,
                -18 * (-1 + s) * s**2 * (-1 + t) * t,
                18 * (-1 + s) * s**2 * (-1 + t) * t,
                -18 * (-1 + s) ** 2 * s * (-1 + t) * t,
            ],
            [
                3 * (-1 + s) ** 2 * (1 + 2 * s) * (-1 + t) * (-1 + 3 * t),
                -3 * s**2 * (-3 + 2 * s) * (1 - 3 * t) * (-1 + t),
                3 * s**2 * (-3 + 2 * s) * t * (-2 + 3 * t),
                -3 * (-1 + s) ** 2 * (1 + 2 * s) * t * (-2 + 3 * t),
            ],
            [
                9 * (-1 + s) ** 2 * s * (-1 + t) * (-1 + 3 * t),
                -9 * (-1 + s) * s**2 * (-1 + t) * (-1 + 3 * t),
                9 * (-1 + s) * s**2 * t * (-2 + 3 * t),
                -9 * (-1 + s) ** 2 * s * t * (-2 + 3 * t),
            ],
        ]
    )


"""
Calculate basis functions at n_sub**2 points
"""


def bf(n_sub):
    # Get the basis functions at each of the points
    lin = np.linspace(0.0, 1.0, n_sub)
    s = np.tensordot(lin, [1] * n_sub, axes=0)
    t = s.transpose()
    return basis_functions(s, t)


def bf_s(n_sub):
    # Get the basis functions at each of the points
    lin = np.linspace(0.0, 1.0, n_sub)
    s = np.tensordot(lin, [1] * n_sub, axes=0)
    t = s.transpose()
    return basis_functions_s(s, t)


def bf_t(n_sub):
    # Get the basis functions at each of the points
    lin = np.linspace(0.0, 1.0, n_sub)
    s = np.tensordot(lin, [1] * n_sub, axes=0)
    t = s.transpose()
    return basis_functions_t(s, t)



"""IMAS-Python version of the paraview plugin classes."""

import getpass
import logging
from abc import ABC, abstractmethod

import imas
import imas.ids_defs
from paraview.util.vtkAlgorithm import smdomain, smhint, smproperty
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from vtkmodules.vtkCommonCore import vtkStringArray

from imas_paraview.convert import Converter
from imas_paraview.imas_uri import uri_from_path, uri_from_pulse_run
from imas_paraview.paraview_support.servermanager_tools import (
    arrayselectiondomain,
    arrayselectionstringvector,
    checkbox,
    doublevector,
    enumeration,
    genericdecorator,
    intvector,
    propertygroup,
    stringlistdomain,
    stringvector,
)
from imas_paraview.util import get_grid_ggd

logger = logging.getLogger("imas_paraview")

if imas.backends.imas_core.imas_interface.has_imas:
    BACKENDS = {
        "MDSplus": imas.ids_defs.MDSPLUS_BACKEND,
        "HDF5": imas.ids_defs.HDF5_BACKEND,
        "ASCII": imas.ids_defs.ASCII_BACKEND,
    }
    """Mapping of UI labels for each backend and ID, used for the Backend dropdown."""
    DEFAULT_BACKEND = imas.ids_defs.MDSPLUS_BACKEND
    """Default backend selected in the UI."""
else:
    BACKENDS = {"[IMAS-Core is not available]": -1}
    DEFAULT_BACKEND = -1


class GGDVTKPluginBase(VTKPythonAlgorithmBase, ABC):
    """GGD Reader based on IMAS-Python"""

    _show_parent_indices_dropdown = False

    def __init_subclass__(cls, use_bezier=False, is_time_dependent=False, **kwargs):
        # Flag to classify time dependent plugins
        cls.is_time_dependent = is_time_dependent
        # Only show bezier methods when the use_bezier flag is enabled.
        bezier_methods = ["PG2_BezierGroup", "P20_SetNPlane", "P21_SetPhiRange"]
        super().__init_subclass__(**kwargs)
        # Copy all function definitions to the subclass
        # This allows paraview to see that they are properties that should be exposed
        for name, value in vars(GGDVTKPluginBase).items():
            if (
                callable(value)
                and not name.startswith("__")
                and not getattr(value, "__isabstractmethod__", False)
                and not name == "request_information"
            ):
                if name in bezier_methods:
                    if use_bezier:
                        setattr(cls, name, value)
                else:
                    setattr(cls, name, value)

    def __init__(self, output_type, supported_ids):
        super().__init__(
            nInputPorts=0,
            nOutputPorts=1,
            outputType=output_type,
        )
        # URI properties
        self._uri_selection_mode = 1
        self._uri_input = ""
        self._uri_path = ""
        self._uri_backend = DEFAULT_BACKEND
        self._uri_pulse = 0
        self._uri_run = 0
        self._uri_database = "ITER"
        self._uri_user = "public"
        self._uri_version = "3"

        # IDS properties
        self._ids_and_occurrence = ""
        self._supported_ids = supported_ids

        # Bezier interpolation properties, the derived classes are responsible for
        # setting self._n_plane to the correct default value
        self._phi_start = 0
        self._phi_end = 0

        # URI is calculated from the possible inputs
        self._uri = ""
        self._uri_error = ""

        # Data caches
        self._dbentry = None
        self._ids_list = []
        self._ids = None

        # Load ggd_idx from paraview UI
        self._time_steps = []
        # Selected parent index
        self._selected_parent_index = 0

        # Values to fill the array selector with
        self._selectable = []
        self._selected = []
        self.show_all = False

        self.converter = None

    def _update_property(self, name, value, callback=None):
        """Convenience method to update a property when value changed."""
        if getattr(self, name) != value:
            setattr(self, name, value)
            if callback is not None:
                callback()
            self.Modified()

    def _calculate_uri(self) -> None:
        """Determine the IMAS URI from user input."""
        if self._uri_selection_mode == 1:
            # Use URI provided by user, removing leading/trailing whitespaces
            uri = self._uri_input
            if uri is not None:
                uri = uri.strip()
        elif self._uri_selection_mode == 2:
            # Determine URI from selected file
            uri = uri_from_path(self._uri_path)
        elif self._uri_selection_mode == 3:
            uri = uri_from_pulse_run(
                self._uri_backend,
                self._uri_database,
                self._uri_pulse,
                self._uri_run,
                self._uri_user,
                self._uri_version,
            )
        else:
            raise ValueError(f"Invalid value for {self._uri_selection_mode=}")

        # Ensure uri is not None:
        uri = uri or ""

        if uri != self._uri:
            self._uri = uri
            if self._dbentry is not None:
                self._dbentry.close()
                self._ids = self._dbentry = None
                self._selectable = []
                self._selected = []
                self._ids_list = []
            self._uri_error = ""
            if self._uri:
                # Try to open the DBEntry
                try:
                    self._dbentry = imas.DBEntry(self._uri, "r")
                    logger.info(f'Successfully opened URI "{self._uri}".')
                except Exception as exc:
                    self._uri_error = str(exc)
            self._update_ids_list()
            self.Modified()

    def _update_ids_list(self) -> None:
        """Update the list of available IDSs in the selected Data Entry."""
        self._ids_list = []
        if self._dbentry is not None:
            for ids_name in self._supported_ids:
                for occurrence in self._dbentry.list_all_occurrences(ids_name):
                    val = ids_name if occurrence == 0 else f"{ids_name}/{occurrence}"
                    self._ids_list.append(val)
            # If dropdown will be used, add "<Select IDS>" dummy value as first entry
            # in list, as this will be taken as default
            if len(self._ids_list) > 1:
                self._ids_list.insert(0, "<Select IDS>")

    # Note on property names:
    # - Properties are sorted in the GUI by name of the function in paraview < 5.13
    # - In paraview 5.13 and newer, this sorting is based on declaration order (see
    #   https://gitlab.kitware.com/paraview/paraview/-/commit/7ea6373ff67c59a7f43b63697aaf6f1ddee51cab)
    # - To properly sort the properties in paraview <= 5.12 we prefix each property with
    #   "P<groupID><propertyID>", e.g. P01_SetURI

    # Properties for setting the URI
    ####################################################################################

    @intvector(name="URISelection", label="", default_values="1")
    @enumeration("enum", {"Enter URI": 1, "Select file": 2, "Enter pulse, run, ..": 3})
    def P00_URISelectionMode(self, mode):
        """Select mode for locating the data entry: manually enter the URI, select a
        local HDF5/MDSplus file, or enter 'legacy' parameters."""
        self._update_property("_uri_selection_mode", mode, self._calculate_uri)

    @stringvector(name="Enter URI", default_values="")
    @genericdecorator(mode="visibility", property="URISelection", value="1")
    def P01_SetURI(self, uri):
        """The IMAS URI for the Data Entry to load."""
        self._update_property("_uri_input", uri, self._calculate_uri)

    @stringvector(name="URI from file", default_values="")
    @smdomain.filelist()
    @smhint.filechooser(
        extensions=["h5", "datafile"],
        file_description="IMAS HDF5/MDSplus backend files",
    )
    @genericdecorator(mode="visibility", property="URISelection", value="2")
    def P02_SetURIFromFileName(self, file):
        """Select a local file from the HDF5 or MDSplus backend."""
        self._update_property("_uri_path", file, self._calculate_uri)

    @intvector(name="Backend", default_values=DEFAULT_BACKEND)
    @enumeration("backend", BACKENDS)
    @genericdecorator(mode="visibility", property="URISelection", value="3")
    def P03_SetBackend(self, backend):
        """Select the IMAS backend that stores the data."""
        self._update_property("_uri_backend", backend, self._calculate_uri)

    @stringvector(name="Database", default_values="")
    @genericdecorator(mode="visibility", property="URISelection", value="3")
    def P04_SetDatabase(self, database):
        """Enter the Database name where the data is stored."""
        self._update_property("_uri_database", database, self._calculate_uri)

    @intvector(name="Pulse", default_values="0")
    @genericdecorator(mode="visibility", property="URISelection", value="3")
    def P05_SetPulse(self, pulse):
        """Pulse number."""
        self._update_property("_uri_pulse", pulse, self._calculate_uri)

    @intvector(name="Run", default_values="0")
    @genericdecorator(mode="visibility", property="URISelection", value="3")
    def P06_SetRun(self, run):
        """Run number."""
        self._update_property("_uri_run", run, self._calculate_uri)

    @stringvector(name="User", default_values=getpass.getuser())
    @genericdecorator(mode="visibility", property="URISelection", value="3")
    def P07_SetUser(self, user):
        """User name for the IMAS database."""
        self._update_property("_uri_user", user, self._calculate_uri)

    @stringvector(name="Version", default_values="3")
    @genericdecorator(mode="visibility", property="URISelection", value="3")
    def P08_SetVersion(self, version):
        """Major version of the DD used to store the data."""
        self._update_property("_uri_version", version, self._calculate_uri)

    @stringvector(name="Status", information_only=1, panel_visibility="default")
    @smhint.xml('<Widget type="one_liner_wrapped" />')
    def P09_GetURIStatus(self):
        """Status information related to the URI."""
        if not self._uri:
            return "No URI selected, press Apply to set URI."
        if self._uri_error:
            return f'Could not open URI "{self._uri}":\n{self._uri_error}'
        num_ids = len(self._ids_list)

        # Ensure <Select IDS> is not counted towards number of IDSs
        if num_ids > 1:
            num_ids -= 1

        return (
            f'Successfully opened URI "{self._uri}".\n'
            f"This Data Entry contains {num_ids} supported IDSs."
        )

    # Properties for setting the IDS name, occurrence and which GGD arrays to load
    ####################################################################################

    def _clear_ids(self):
        """Helper callback to clear any loaded IDS when idsname/occurrence change."""
        self._ids = None

    @stringvector(name="IDSAndOccurrence", label="IDS/Occurrence")
    @stringlistdomain("IDSList", name="ids_list")
    def P10_SetIDSAndOccurrence(self, value):
        """Select the IDS/occurrence to load.

        The dropdown is disabled when no Data Entry has been loaded (press Apply after
        selecting a URI to load the Data Entry), or when the loaded Data Entry contains
        no IDSs supported by this plugin.
        """
        if value == "<Select IDS>":
            self._selectable = []
            self._selected = []
            value = ""
        self._update_property("_ids_and_occurrence", value, self._clear_ids)

    @stringvector(name="IDSList", information_only=1, si_class="vtkSIDataArrayProperty")
    def P11_GetIDSList(self):
        """Return a list of IDSs with data inside the selected Data Entry."""
        arr = vtkStringArray()
        for val in self._ids_list:
            arr.InsertNextValue(val)
        return arr

    @stringvector(
        name="ParentIndices", information_only=1, si_class="vtkSIDataArrayProperty"
    )
    def P11_GetParentIndices(self):
        """Return a list of parent indices, if applicable for the selected IDS."""
        arr = vtkStringArray()
        if self._show_parent_indices_dropdown and self._ids:
            try:
                grid_ggd = get_grid_ggd(self._ids)
            except Exception:
                logger.error("Could not get parent indices", exc_info=1)
                return arr
            # Try to find any non-time-dependent indices in the parents
            node = grid_ggd
            while node:
                parent = imas.util.get_parent(node)
                # if parent and node have the same metadata, it means node = parent[idx]
                if parent and parent.metadata is node.metadata:
                    if not node.metadata.coordinate1.is_time_coordinate:
                        for i in range(len(parent)):
                            arr.InsertNextValue(f"{node.metadata.name}[{i}]")
                node = parent
        if arr.GetNumberOfValues() == 0:
            arr.InsertNextValue("N/A")
        return arr

    @stringvector(name="ParentIndex", label="Sub index")
    @stringlistdomain("ParentIndices", name="parent_indices")
    @genericdecorator(mode="visibility", property="ParentIndexVisible", value="1")
    def P11_SetParentIndex(self, value):
        """Select structure index to extract the GGD grid from. This is only applicable
        to some IDSs, such as 'wall'."""
        # Extract index from the selected text
        index = 0
        if "[" in value:
            index = int(value[value.rfind("[") + 1 : -1])
        self._update_property("_selected_parent_index", index)

    @intvector(name="ParentIndexVisible", information_only=1, panel_visibility="never")
    def P11_GetParentIndexVisible(self):
        # Only show the Sub index dropdown (previous method) for the GGDReader
        return int(self._show_parent_indices_dropdown)

    @arrayselectiondomain(
        property_name="AttributeArray",
        name="AttributeArraySelector",
        label="Select attribute Arrays",
    )
    def P12_SetAttributeArray(self, array, status):
        """Select all or a subset of available GGD arrays to load."""
        # Add an array to selected list
        if status == 1 and array not in self._selected:
            self._selected.append(array)
            self.Modified()

        # Remove an array from selected list
        if status == 0 and array in self._selected:
            self._selected.remove(array)
            self.Modified()

    @arrayselectionstringvector(
        property_name="AttributeArray", attribute_name="Attribute"
    )
    def _AttributeArraySelector(self):
        pass

    def GetNumberOfAttributeArrays(self):
        return len(self._selectable)

    @abstractmethod
    def GetAttributeArrayName(self, idx) -> str:
        """Retrieve the name of the attribute array at a specified index.
        This method should be implemented by subclasses to return the name
        of an attribute array based on its index.

        Args:
            idx: The index of the attribute array.

        Returns:
            The name of the attribute array corresponding to the given index.
        """
        pass

    def GetAttributeArrayStatus(self, *args):
        return 1

    @checkbox(
        name="ShowAll",
        label="Show All Attribute Arrays",
        default_values="0",
    )
    def P13_SetShowEmpty(self, val):
        """Enable to show all possible attributes, even when they are empty. When
        disabled, only the the attributes in the selector window which are filled
        in the first time step are shown."""
        if val == 1:
            self.show_all = True
            logger.info(
                "Showing all attributes in selector window, including empty ones"
            )
        else:
            self.show_all = False
            logger.info(
                "Showing only the attributes that are filled in the first time "
                "step in selector window."
            )
        self.Modified()

    @checkbox(
        name="LazyLoading",
        label="Preload Data",
        default_values="0",
    )
    def P14_SetLazyLoading(self, val):
        """Turn on to preload the entire IDS beforehand, if this is left off, the data
        is loaded on-demand through lazy loading. It is recommended to leave this off
        if you are only loading a small subset of the data. If you want to load most of
        the data, it is recommended to turn this on."""
        if val == 0:
            status = "enabled"
            self.lazy = True
        else:
            status = "disabled"
            self.lazy = False
        logger.info(f"Lazy Loading is {status}.")
        self.Modified()

    # Properties for handling time steps
    ####################################################################################

    @smproperty.doublevector(
        name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty"
    )
    def GetTimestepValues(self):
        return self._time_steps

    # Properties for Bezier interpolation
    ####################################################################################

    @intvector(name="N plane", default_values=1)
    def P20_SetNPlane(self, val):
        self._update_property("_n_plane", val)

    @doublevector(name="Phi range", default_values=[0, 0])
    @smdomain.doublerange(min=0, max=360.0)
    def P21_SetPhiRange(self, val, val2):
        self._update_property("_phi_start", val)
        self._update_property("_phi_end", val2)

    # Property groups: sorted by name and must be alphabetically after the properties
    ####################################################################################

    @propertygroup(
        "Data entry URI",
        [
            "URISelection",
            "Enter URI",
            "URI from file",
            "Backend",
            "Database",
            "Pulse",
            "Run",
            "User",
            "Version",
            "Status",
        ],
    )
    def PG0_DataEntryGroup(self):
        """Dummy function to define a PropertyGroup."""

    @propertygroup(
        "Select IDS",
        [
            "IDSAndOccurrence",
            "IDSList",
            "ParentIndices",
            "ParentIndex",
            "AttributeArraySelector",
            "ShowAll",
            "LazyLoading",
        ],
    )
    def PG1_IDSGroup(self):
        """Dummy function to define a PropertyGroup."""

    @propertygroup("Bezier interpolation settings", ["N plane", "Phi range"])
    def PG2_BezierGroup(self):
        """Dummy function to define a PropertyGroup."""

    # Implement VTK algorithm
    ####################################################################################

    def RequestInformation(self, request, inInfo, outInfo):
        if len(self._ids_list) == 1:
            value = self._ids_list[0]
            self._update_property("_ids_and_occurrence", value, self._clear_ids)

        if self._dbentry is None or not self._ids_and_occurrence:
            return 1

        # Load IDS and available time steps
        idsname, _, _ = self._ids_and_occurrence.partition("/")
        if idsname not in self._ids_list:
            logger.warning("Could not find the selected IDS.")
            self._selectable = []
            self._selected = []
            return 1

        self._load_ids_from_backend()
        self.request_information()

        # TODO: Add support for IDSs with heterogeneous time mode
        if self.is_time_dependent:
            if (
                self._ids.ids_properties.homogeneous_time
                == imas.ids_defs.IDS_TIME_MODE_HETEROGENEOUS
            ):
                logger.error("Heterogeneous IDSs are currently not supported.")
                return 1
            self._time_steps = self._ids.time

            # Pass time steps to Paraview
            executive = self.GetExecutive()
            outInfo = outInfo.GetInformationObject(0)
            outInfo.Remove(executive.TIME_STEPS())
            outInfo.Remove(executive.TIME_RANGE())

            for time_step in self._time_steps:
                outInfo.Append(executive.TIME_STEPS(), time_step)
            outInfo.Append(executive.TIME_RANGE(), self._time_steps[0])
            outInfo.Append(executive.TIME_RANGE(), self._time_steps[-1])
        return 1

    def _load_ids_from_backend(self):
        """If the IDS is None, it will load the selected IDS from the backend. If lazy
        loading is disabled, the entire IDS will be loaded during this step.
        """
        if self._ids is None:
            idsname, _, occurrence = self._ids_and_occurrence.partition("/")
            occurrence = int(occurrence or 0)
            logger.info("Loading IDS %s/%d ...", idsname, occurrence)

            if self._dbentry is None:
                raise RuntimeError(
                    "Could not fetch IDS from the database entry, as the entry is empty"
                )
            self._ids = self._dbentry.get(
                idsname,
                occurrence,
                autoconvert=False,
                lazy=self.lazy,
                ignore_unknown_dd_version=True,
            )
            self.setup_ids()
            self.converter = Converter(self._ids, dbentry=self._dbentry)

    def request_information(self):
        """
        Called during the RequestInformation stage in ParaView. Subclasses can implement
        this method to define actions that need to occur during the RequestInformation
        stage of the Paraview pipeline.
        """
        pass

    @abstractmethod
    def setup_ids(self):
        """
        Called after an IDS is loaded for the first time. Subclasses should implement
        this method to define any initialization or setup processes that should only
        run once after the IDS is loaded.
        """
        pass

    def _get_selected_time_step(self, outInfo):
        """Retrieves the selected time step based on the time selection widget in
        the ParaView UI.

        Args:
            outInfo: An information object that holds details about the update request

        Returns:
            time_step: The selected time step. If the selected time step cannot be found
            in the IDS, returns None.
        """
        # Retrieve time step from time selection widget in Paraview UI
        executive = self.GetExecutive()
        outInfo = outInfo.GetInformationObject(0)
        time_step = outInfo.Get(executive.UPDATE_TIME_STEP())

        # Paraview (v5.12.1) contains a bug where, if you load a dataset with only a
        # single timestep, it automatically loads 10 synthetic timesteps ranging from
        # timestep upto time step + 1. The issue can be found here:
        # https://gitlab.kitware.com/paraview/paraview/-/issues/22360
        # A workaround is as follows:
        # 1. load the single timestep data
        # 2. open "View > Time Manager"
        # 3. uncheck and check again the checkbox "Time Sources"
        # For now, return None if such a timestep is chosen

        # Check if it exists in list of time steps
        if time_step in self._time_steps:
            logger.debug(f"Selected time step in Paraview: {time_step}")
            return time_step
        elif self._time_steps:
            logger.info(
                f"Selected time step {time_step} was not found in the IDS. "
                f"The first time step ({self._time_steps[0]}) is loaded instead."
            )
            return self._time_steps[0]
        else:
            return None



def profile(fname):
    """Decorator to cProfile the code inside the function.

    Args:
        fname: Filename to dump the cProfile statistics to.

    Example:
        .. code-block:: python

            class Test(VTKPythonAlgorithmBase):
                @profile("RequestData.stats")
                def RequestData(self, request, inInfo, outInfo):
                    ...
    """
    import cProfile
    import functools

    def inner(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            prof = cProfile.Profile()
            try:
                prof.enable()
                return func(*args, **kwargs)
            finally:
                prof.disable()
                prof.dump_stats(fname)
                print("Statistics written to", fname)

        return wrapper

    return inner



import pytest
from imas.ids_defs import ASCII_BACKEND, HDF5_BACKEND, MDSPLUS_BACKEND

from imas_paraview.imas_uri import uri_from_path, uri_from_pulse_run


def test_uri_from_path():
    # Test a couple of invalid paths
    assert uri_from_path(None) is None
    assert uri_from_path("") is None
    assert uri_from_path("xyz.test") is None
    assert uri_from_path("xyz") is None
    # Valid HDF5:
    assert uri_from_path("test/a.h5") == "imas:hdf5?path=test"
    assert uri_from_path("test/master.h5") == "imas:hdf5?path=test"
    # Valid MDSplus
    assert uri_from_path("test/ids_001.characteristics") == "imas:mdsplus?path=test"
    assert uri_from_path("test/ids_001.datafile") == "imas:mdsplus?path=test"
    assert uri_from_path("test/ids_001.tree") == "imas:mdsplus?path=test"


@pytest.mark.skip(reason="no IMAS-Core available")
def test_uri_from_pulse_run():
    # Missing values result in a None uri
    assert uri_from_pulse_run(-1, "a", 1, 2, "b", "c") is None
    assert uri_from_pulse_run(HDF5_BACKEND, "", 1, 2, "b", "c") is None
    assert uri_from_pulse_run(HDF5_BACKEND, "a", 1, 2, "", "c") is None
    assert uri_from_pulse_run(HDF5_BACKEND, "a", 1, 2, "b", "") is None

    # Valid parameters:
    assert (
        uri_from_pulse_run(HDF5_BACKEND, "db", 1, 2, "user", "3")
        == "imas:hdf5?user=user;pulse=1;run=2;database=db;version=3"
    )
    assert (
        uri_from_pulse_run(MDSPLUS_BACKEND, "db", 1, 2, "user", "4")
        == "imas:mdsplus?user=user;pulse=1;run=2;database=db;version=4"
    )
    assert (
        uri_from_pulse_run(ASCII_BACKEND, "db", 1, 2, "user", "3")
        == "imas:ascii?user=user;pulse=1;run=2;database=db;version=3"
    )



import logging

import imas
from packaging import version

from imas_paraview.vtkhandler import VTKHandler

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

# Setup logger
logger = logging.getLogger("imas_paraview")
handler = VTKHandler()
logger.addHandler(handler)
logger.setLevel(handler.get_level())

# Remove imas handler and add the VTK handler to make sure IMAS-Python logs are
# correctly displayed in Paraview's logs
imas_logger = logging.getLogger("imas")
if len(imas_logger.handlers) == 1:
    imas_logger.removeHandler(imas_logger.handlers[0])
imas_logger.addHandler(handler)
imas_logger.setLevel(handler.get_level())

# Check IMAS-Python version
if not hasattr(imas, "__version__"):
    raise RuntimeError(
        """
[ERROR] Detected an outdated version of the 'imas' module.

The installed 'imas' package appears to be an incompatible legacy
version of the high-level Python interface of the IMAS Access Layer.

To resolve this, remove / unload this version and re-install using:

    pip install imas-python

or load the appropriate environment module on your system, e.g.

    module load IMAS-Python

More info: https://pypi.org/project/imas-python/
"""
    )
imas_version = imas.__version__
required_imas_python_version = "2.0.0"
if version.parse(imas_version) < version.parse(required_imas_python_version):
    logger.warning(
        f"IMAS-Python version {imas_version} is lower than the recommended version "
        f"{required_imas_python_version}. Some features might not work as expected."
    )



import logging

from vtkmodules.vtkCommonDataModel import vtkDataObject, vtkPartitionedDataSetCollection

from imas_paraview.convert import InterpSettings
from imas_paraview.io import read_ps
from imas_paraview.plugins.base_class import GGDVTKPluginBase
from imas_paraview.progress import Progress

logger = logging.getLogger("imas_paraview")


class GGDBaseReader(GGDVTKPluginBase):
    """Base class for GGDReader and the JorekGGDReader."""

    def __init__(self, supported_ids):
        super().__init__("vtkPartitionedDataSetCollection", supported_ids)
        # GGD arrays to load
        self._all_vector_paths = []
        self._all_scalar_paths = []

    def GetAttributeArrayName(self, idx) -> str:
        return self._name_from_idspath(self._selectable[idx])

    def RequestDataObject(self, request, inInfo, outInfo):
        output = vtkPartitionedDataSetCollection()
        outInfo.GetInformationObject(0).Set(vtkDataObject.DATA_OBJECT(), output)
        outInfo.GetInformationObject(0).Set(
            vtkDataObject.DATA_EXTENT_TYPE(), output.GetExtentType()
        )
        return 1

    def RequestData(self, request, inInfo, outInfo):
        if self._dbentry is None or not self._ids_and_occurrence or self._ids is None:
            return 1

        # Retrieve the selected time step and GGD arrays
        selected_scalar_paths, selected_vector_paths = self._get_selected_ggd_paths()
        time = self._get_selected_time_step(outInfo)
        if time is None:
            logger.warning("Selected invalid time step")
            return 1

        # Create progress object to advance Paraview progress bar
        progress = Progress(self.UpdateProgress)

        if self._n_plane < self._MIN_N_PLANE:
            raise ValueError(
                f"The number of Bezier planes cannot be less than {self._MIN_N_PLANE}."
            )

        plane_config = InterpSettings(
            n_plane=self._n_plane, phi_start=self._phi_start, phi_end=self._phi_end
        )
        output = self.converter.ggd_to_vtk(
            time=time,
            scalar_paths=selected_scalar_paths,
            vector_paths=selected_vector_paths,
            plane_config=plane_config,
            outInfo=outInfo,
            progress=progress,
            parent_idx=self._selected_parent_index,
        )

        if output is None:
            logger.warning("Could not convert GGD to VTK.")
        return 1

    def request_information(self):
        """
        Select which GGD arrays to show in the array domain selector, based
        on whether the "Show All" checkbox is enabled.
        """
        if self._ids is not None:
            if self.show_all:
                self._selectable = self._all_vector_paths + self._all_scalar_paths
            else:
                self._selectable = self._filled_vector_paths + self._filled_scalar_paths
        pass

    def setup_ids(self):
        """
        Initializes plasma state reader and populates scalar and vector paths for
        the array selection domaindomain selection.
        """
        if self._ids is not None:
            # Load paths from IDS
            ps_reader = read_ps.PlasmaStateReader(self._ids)
            (
                self._all_scalar_paths,
                self._all_vector_paths,
                self._filled_scalar_paths,
                self._filled_vector_paths,
            ) = ps_reader.load_paths_from_ids()

    def _name_from_idspath(self, path):
        """Converts an IDSPath to a string by removing 'ggd' and capitalizing each part
        of the path, with the parts separated by spaces.

        Example:
            If path is IDSPath('ggd/electrons/pressure'), the function returns
            "Electrons Pressure"

        Args:
            path: The IDSPath object to convert into a formatted string

        Returns:
            A formatted string of the IDSPath
        """
        path_list = list(path.parts)
        if "ggd" in path_list:
            path_list.remove("ggd")
        if "description_ggd" in path_list:
            path_list.remove("description_ggd")
        for i in range(len(path_list)):
            path_list[i] = path_list[i].capitalize()

        name = " ".join(path_list)

        # If GGDs are not filled in the first time step, add (?) to their name. This
        # notifies that the GGD array is not filled and will most likely not contain
        # any data. We do not remove these GGD arrays from the selector window entirely,
        # because later time steps could still contain data.
        if (
            path not in self._filled_scalar_paths
            and path not in self._filled_vector_paths
        ):
            name = f"{name} (?)"
        return name

    def _get_selected_ggd_paths(self):
        """Retrieve the IDSPaths of the selected scalar and vector GGD arrays.

        Returns:
            selected_scalar_paths: List of IDSPaths of selected scalar GGD arrays
            selected_vector_paths: List of IDSPaths of selected vector GGD arrays
        """

        # Determine if selected GGD arrays are scalar or vector arrays
        selected_scalar_paths = [
            obj
            for obj in self._all_scalar_paths
            if self._name_from_idspath(obj) in self._selected
        ]
        selected_vector_paths = [
            obj
            for obj in self._all_vector_paths
            if self._name_from_idspath(obj) in self._selected
        ]
        return selected_scalar_paths, selected_vector_paths



from imas.ids_struct_array import IDSStructArray


def _get_nodes_from_path(
    node, path, get_empty_arrays, ggd_idx=None, create_empty_structs=False
):
    """Retrieve a list of nodes from a given IDSPath.

    Args:
        node: The starting node to navigate.
        path: An IDSPath to traverse.
        get_empty_arrays (bool): Whether to return empty GGD arrays
        ggd_idx: The GGD time step to load. Defaults to None, which corresponds with
        loading all timesteps.
        create_empty_structs: If required to traverse through an empty structure to
            reach a GGD array, resize this structure to have length 1.

    Returns:
        A list of nodes obtained from the specified path.
    """
    return list(
        _iter_nodes_from_path(
            node, path.parts, get_empty_arrays, ggd_idx, create_empty_structs
        )
    )


def _iter_nodes_from_path(
    node, path_parts, get_empty_arrays, ggd_idx, create_empty_structs
):
    """Recursively iterate through nodes of an IDS node based on path parts.

    Args:
        node: The current node being traversed.
        path_parts: A list of IDSPath segments (parts).
        get_empty_arrays (bool): Whether to return empty GGD arrays
        ggd_idx: The GGD time step to load.
        create_empty_structs: If required to traverse through an empty structure to
            reach a GGD array, resize this structure to have length 1.

    Yields:
        The next node in the structure corresponding to the current path part.
    """
    child_node = node[path_parts[0]]
    if len(path_parts) == 1:
        # The path_parts refer to nodes that have a defined length, such as struct
        # arrays
        if len(child_node) >= 1 or get_empty_arrays:
            yield child_node
    elif isinstance(child_node, IDSStructArray):
        # Only load specific timeidx from ggd node
        if ggd_idx is not None and path_parts[0] == "ggd":
            if len(child_node) > ggd_idx:
                structure = child_node[ggd_idx]
                yield from _iter_nodes_from_path(
                    structure,
                    path_parts[1:],
                    get_empty_arrays,
                    ggd_idx,
                    create_empty_structs,
                )
        else:
            # If the structure is empty, we cannot traverse to the GGD arrays which are
            # (grand)children of this node. Therefore, we need to resize it.
            if len(child_node) == 0 and create_empty_structs:
                child_node.resize(1)

            for structure in child_node:
                yield from _iter_nodes_from_path(
                    structure,
                    path_parts[1:],
                    get_empty_arrays,
                    ggd_idx,
                    create_empty_structs,
                )
    else:
        yield from _iter_nodes_from_path(
            child_node, path_parts[1:], get_empty_arrays, ggd_idx, create_empty_structs
        )



from packaging.version import Version

import imas_paraview


def test_version():
    version = imas_paraview.__version__
    assert Version(version) > Version("0")
    assert "unknown" not in version



class Progress:
    def __init__(self, update_progress):
        self.update_progress = update_progress
        self.progress = 0

    def increment(self, increment):
        """Increments the progress by a specified amount.

        Args:
            increment: amount to increment the progress by.
        """
        self.progress = self.progress + increment
        self.update_progress(self.progress)

    def set(self, value):
        """Sets the progress to a specific value (0 to 1).

        Args:
            value: specific value to set the progress to
        """
        self.progress = value
        self.update_progress(value)



import logging

from paraview.util.vtkAlgorithm import smhint, smproxy

from imas_paraview.plugins.ggd_base_reader import GGDBaseReader

logger = logging.getLogger("imas_paraview")


SUPPORTED_IDS_NAMES = [
    "edge_profiles",
    "edge_sources",
    "edge_transport",
    "mhd",
    "radiation",
    "runaway_electrons",
    "wall",
    "plasma_profiles",
    "plasma_sources",
    "plasma_transport",
]

# FIXME: Some IDSs do not have the grid structure in a separate `grid_ggd` object, as
# described by the GGD guidelines. They are for now denoted as experimental IDS, as they
# are currently not covered by the unit tests. If the DD for these stays like this, they
# will need to be handled separately. GGD grid locations for each IDS:
# equilibrium (grids_ggd/grid)
# distribution_sources (source/ggd/grid)
# distributions (distribution/ggd/grid)
# tf (field_map/grid)
# transport_solver_numerics (boundary_conditions_ggd/grid)
# waves (coherent_wave/full_wave/grid)

EXPERIMENTAL_IDS_NAMES = [
    "equilibrium",
    "distribution_sources",
    "distributions",
    "tf",
    "transport_solver_numerics",
    "waves",
]


@smproxy.source(label="GGD Reader")
@smhint.xml("""<ShowInMenu category="IMAS Tools" />""")
class GGDReader(GGDBaseReader, is_time_dependent=True):
    _show_parent_indices_dropdown = True

    def __init__(self):
        supported_ids = EXPERIMENTAL_IDS_NAMES + SUPPORTED_IDS_NAMES
        super().__init__(supported_ids)
        self._MIN_N_PLANE = self._n_plane = 0



import numpy as np
import pytest

from imas_paraview.util import find_closest_indices


@pytest.mark.parametrize(
    "request_times, expected_indices",
    [
        ([2.2, 3.3], [1, 2]),
        ([4.4, 5.5], [3, 4]),
        ([1.5, 3.8, 5.9], [0, 2, 4]),
        ([10], [4]),
        ([0.1, 0.5, 0.9, 1.01], []),
        ([0.5, 1.01, 2.5, 4.7], [1, 3]),
        ([2.2, 2.2, 4.4, 4.4], [1, 1, 3, 3]),
    ],
)
def test_find_closest_indices(request_times, expected_indices):
    time_array = np.array([1.1, 2.2, 3.3, 4.4, 5.5])
    time_indices = find_closest_indices(request_times, time_array)
    assert time_indices == expected_indices



import imas
import pytest
from imas.ids_path import IDSPath

from imas_paraview.ids_util import (
    create_name_recursive,
    get_arrays_from_ids,
    recursive_ggd_path_search,
)
from imas_paraview.tests.fill_ggd import fill_scalar_quantity, fill_vector_quantity


@pytest.fixture
def ids_fixture():
    """Creates a dummy IDS with 4 time steps"""
    ids = imas.IDSFactory().new("edge_sources")
    ids.time = [1.1, 2.2, 3.3, 4.4]
    num_timesteps = len(ids.time)
    ids.source.resize(1)
    ids.source[0].ggd.resize(num_timesteps)
    input_scalars = [None] * num_timesteps
    input_vectors = [None] * num_timesteps
    for ggd_idx in range(num_timesteps):
        ids.source[0].ggd[ggd_idx].ion.resize(1)
        ids.source[0].ggd[ggd_idx].ion[0].state.resize(1)
        input_scalars[ggd_idx] = ids.source[0].ggd[ggd_idx].ion[0].energy
        input_vectors[ggd_idx] = ids.source[0].ggd[ggd_idx].ion[0].momentum
    return ids, input_scalars, input_vectors


def fill_quantities(ids, input_scalars, input_vectors):
    """Fills the input scalar and vector GGD arrays with dummy values"""
    num_vertices = 6
    num_edges = 7
    num_faces = 2
    for ggd_idx in range(len(ids.time)):
        fill_scalar_quantity(input_scalars[ggd_idx], num_vertices, num_edges, num_faces)
        fill_vector_quantity(input_vectors[ggd_idx], num_vertices, num_edges, num_faces)


def test_get_arrays_from_ids_empty(ids_fixture):
    """Tests get_arrays_from_ids when IDS has no filled GGD arrays"""
    ids, _, _ = ids_fixture
    retrieved_scalar_arrays, retrieved_vector_arrays = get_arrays_from_ids(ids)
    assert retrieved_scalar_arrays == []
    assert retrieved_vector_arrays == []


def test_get_arrays_from_ids_all_timesteps(ids_fixture):
    """Tests get_arrays_from_ids when requesting all time steps in the IDS"""

    ids, input_scalars, input_vectors = ids_fixture
    fill_quantities(ids, input_scalars, input_vectors)

    retrieved_scalar_arrays, retrieved_vector_arrays = get_arrays_from_ids(ids)
    assert len(retrieved_scalar_arrays) == len(ids.time)
    assert len(retrieved_vector_arrays) == len(ids.time)
    assert all(scalar in retrieved_scalar_arrays for scalar in input_scalars)
    assert all(vector in retrieved_vector_arrays for vector in input_vectors)


def test_get_arrays_from_ids_single_timesteps(ids_fixture):
    """Tests get_arrays_from_ids when requesting only a single time step in the IDS"""

    ids, input_scalars, input_vectors = ids_fixture
    fill_quantities(ids, input_scalars, input_vectors)

    for ggd_idx in range(len(ids.time)):
        retrieved_scalar_arrays, retrieved_vector_arrays = get_arrays_from_ids(
            ids, ggd_idx=ggd_idx
        )
        assert len(retrieved_scalar_arrays) == 1
        assert len(retrieved_vector_arrays) == 1
        assert input_scalars[ggd_idx] == retrieved_scalar_arrays[0]
        assert input_vectors[ggd_idx] == retrieved_vector_arrays[0]


def test_get_arrays_from_ids_not_defined_timesteps(ids_fixture):
    """Tests get_arrays_from_ids when requesting a single time step that does not exist
    in the IDS"""
    ids, input_scalars, input_vectors = ids_fixture
    fill_quantities(ids, input_scalars, input_vectors)

    # Get timestep not in IDS
    ggd_idx = len(ids.time)
    retrieved_scalar_arrays, retrieved_vector_arrays = get_arrays_from_ids(
        ids, ggd_idx=ggd_idx
    )
    assert retrieved_scalar_arrays == []
    assert retrieved_vector_arrays == []


def test_recursive_ggd_path_search():
    """Test if recursive_ggd_path_search returns all GGD array paths"""
    ids = imas.IDSFactory(version="3.41.0").new("edge_sources")

    es_scalar_v3_41 = [
        IDSPath("source/ggd/electrons/particles"),
        IDSPath("source/ggd/electrons/energy"),
        IDSPath("source/ggd/ion/energy"),
        IDSPath("source/ggd/ion/particles"),
        IDSPath("source/ggd/ion/state/particles"),
        IDSPath("source/ggd/ion/state/energy"),
        IDSPath("source/ggd/neutral/particles"),
        IDSPath("source/ggd/neutral/energy"),
        IDSPath("source/ggd/neutral/state/particles"),
        IDSPath("source/ggd/neutral/state/energy"),
        IDSPath("source/ggd/total_ion_energy"),
        IDSPath("source/ggd/current"),
    ]
    es_vector_v3_41 = [
        IDSPath("source/ggd/ion/momentum"),
        IDSPath("source/ggd/ion/state/momentum"),
        IDSPath("source/ggd/neutral/momentum"),
        IDSPath("source/ggd/neutral/state/momentum"),
        IDSPath("source/ggd/momentum"),
    ]

    scalar_arrays = []
    vector_arrays = []
    recursive_ggd_path_search(ids.metadata, scalar_arrays, vector_arrays)
    assert set(scalar_arrays) == set(es_scalar_v3_41)
    assert set(vector_arrays) == set(es_vector_v3_41)


def test_create_name_recursive_ggd():
    ids = imas.IDSFactory(version="3.40.0").new("edge_sources")

    # Resize relevant structures
    ids.source.resize(1)
    ids.source[0].ggd.resize(1)
    ids.source[0].ggd[0].ion.resize(1)
    ids.source[0].ggd[0].ion[0].state.resize(1)
    array = ids.source[0].ggd[0].ion[0].state[0].energy

    # Set identifier name and labels
    ids.source[0].identifier.name = "Charge exchange"
    ids.source[0].ggd[0].ion[0].label = "Ar"
    ids.source[0].ggd[0].ion[0].state[0].label = "Ar+14"

    name = create_name_recursive(array)
    assert name == "Source (Charge exchange) Ion (Ar) State (Ar+14) Energy"


def test_create_name_recursive_profiles():
    ids = imas.IDSFactory(version="3.40.0").new("core_profiles")

    ids.profiles_1d.resize(1)
    phi_potential = ids.profiles_1d[0].phi_potential
    electrons_pressure = ids.profiles_1d[0].electrons.pressure
    ids.profiles_1d[0].ion.resize(2)
    ids.profiles_1d[0].ion[0].label = "D"
    temperature = ids.profiles_1d[0].ion[0].temperature
    ids.profiles_1d[0].ion[1].label = "T"
    density = ids.profiles_1d[0].ion[1].density

    assert create_name_recursive(phi_potential) == "Phi_potential"
    assert create_name_recursive(electrons_pressure) == "Electrons Pressure"
    assert create_name_recursive(temperature) == "Ion (D) Temperature"
    assert create_name_recursive(density) == "Ion (T) Density"



import logging

import numpy as np
from imas.ids_data_type import IDSDataType
from imas.ids_struct_array import IDSStructArray
from vtkmodules.numpy_interface import dataset_adapter as dsa
from vtkmodules.vtkCommonCore import vtkDoubleArray
from vtkmodules.vtkCommonDataModel import vtkCellData, vtkPointData, vtkUnstructuredGrid

from imas_paraview.ids_util import (
    create_name_recursive,
    get_arrays_from_ids,
    recursive_ggd_path_search,
)

# We'll need these below when we create some units manually:
from imas_paraview.util import format_units

logger = logging.getLogger("imas_paraview")
u_pre = "["
u_post = "]"


class PlasmaStateReader:
    def __init__(self, ids):
        """Initializes plasma state reader and retrieves all filled GGD scalar and
        vector arrays from the IDS.

        Args:
            ids: The IDS to load GGD arrays from
        """

        # _cache stores names for each node to avoid recomputing them. It checks if
        # a node's name is already cached before generating it, speeding up the process
        # and ensuring names are computed only once.
        self._cache = {}
        self._ids = ids
        self.scalar_array_list = []
        self.vector_array_list = []

    def load_paths_from_ids(self, ggd_idx=0, return_empty=False):
        """Retrieves scalar and vector array paths from the IDS metadata by performing a
        recursive search through GGD paths.

        Args:
            ggd_idx: Retrieve filled paths of a specific time step.
            return_empty: Whether to return the paths of empty GGD arrays. Defaults to
                False.

        Returns:
            scalar_array_paths: A list of paths of GGD scalar arrays in the IDS
            vector_array_paths: A list of paths of GGD vector arrays in the IDS
        """

        logger.debug("Retrieving GGD paths from IDS metadata")
        all_scalar_paths = []
        all_vector_paths = []
        filled_scalar_paths = []
        filled_vector_paths = []
        recursive_ggd_path_search(
            self._ids.metadata, all_scalar_paths, all_vector_paths
        )

        if return_empty:
            logger.debug("Retrieving all GGD arrays, including empty ones.")
        else:
            logger.debug("Only retrieving filled GGD arrays.")
            filled_scalar_paths = self._get_filled_arrays(
                all_scalar_paths, ggd_idx=ggd_idx
            )
            filled_vector_paths = self._get_filled_arrays(
                all_vector_paths, ggd_idx=ggd_idx
            )

        return (
            all_scalar_paths,
            all_vector_paths,
            filled_scalar_paths,
            filled_vector_paths,
        )

    def load_arrays_from_path(self, ggd_idx, scalar_array_paths, vector_array_paths):
        """Retrieves scalar and vector arrays from the IDS based on the provided time
        index and array paths, and stores them in the respective lists.

        Args:
            ggd_idx: The time index for which to load GGD arrays
            scalar_array_paths: A list of paths of GGD scalar arrays in the IDS
            vector_array_paths: A list of paths of GGD vector arrays in the IDS
        """
        logger.debug("Retrieving GGD arrays from IDS")
        self.scalar_array_list, self.vector_array_list = get_arrays_from_ids(
            self._ids,
            ggd_idx=ggd_idx,
            scalar_array_paths=scalar_array_paths,
            vector_array_paths=vector_array_paths,
        )
        logger.debug(
            f"Found {len(self.scalar_array_list)} scalar arrays and "
            f"{len(self.vector_array_list)} vector arrays in the IDS."
        )

    def read_plasma_state(self, subset_idx: int, ugrid: vtkUnstructuredGrid) -> None:
        """Reads plasma state data arrays from the ggd node in the IDS. These arrays are
        added as point data or cell data to the unstructured grid.

        Args:
            subset_idx: an index into grid_ggd/grid_subset AoS
            ugrid: the unstructured grid instance
        """
        # TODO: GGD-fast

        # Read scalar arrays
        logger.debug("Converting scalar GGD arrays to VTK field data")
        for scalar_array in self.scalar_array_list:
            name = self._create_name_with_units(scalar_array)
            self._add_aos_scalar_array_to_vtk_field_data(
                scalar_array, subset_idx, name, ugrid
            )

        # Read vector arrays
        logger.debug("Converting Vector GGD arrays to VTK field data")
        for vector_array in self.vector_array_list:
            name = self._create_name_with_units(vector_array)
            self._add_aos_vector_array_to_vtk_field_data(
                vector_array, subset_idx, name, ugrid
            )

    def _get_filled_arrays(self, path_list, ggd_idx):
        """Look through a list of IDSPaths containing GGD arrays and return a list
        with only the IDSPaths of GGD arrays which are filled.

        Args:
            path_list: A list of IDSPaths containing GGD arrays.
            ggd_idx: Index for selecting specific GGD array.

        Returns:
            A list containing only the paths of filled GGD arrays.
        """
        filled_arrays = []
        for path in path_list:
            if self._is_path_filled(path, ggd_idx):
                filled_arrays.append(path)
        return filled_arrays

    def _is_path_filled(self, path, ggd_idx):
        """Checks if the provided GGD path for a given time index is filled.

        Args:
            path: IDSPath containing a GGD array.
            ggd_idx: Index for selecting specific GGD array.
        """

        def _check_subnode(node, parts, idx):
            """
            Recursive helper function
            """

            if idx == len(parts):
                return node.size > 0
            part = parts[idx]
            if isinstance(node, IDSStructArray):
                name = getattr(node.metadata, "name", None)
                if name == "ggd":
                    if len(node) <= ggd_idx:
                        return False
                    node = node[ggd_idx]
                    return _check_subnode(node, parts, idx)

                elif len(node) >= 1:
                    for subnode in node:
                        if _check_subnode(subnode, parts, idx):
                            return True
            else:
                node = node[part]
                return _check_subnode(node, parts, idx + 1)

        node = self._ids[path.parts[0]]
        return _check_subnode(node, path.parts, 1)

    def _create_name_with_units(self, array):
        """Creates a name for the GGD array based on its path and units.

        Args:
            array: The ggd scalar or vector array to create the name for

        Returns:
            name_with_units: The name and units of the provided GGD array
        """

        # Format the path in a neat format
        name = self._create_ggd_name(array)

        # Get units for this quantity
        units = format_units(array)

        # Combine name and units
        name_with_units = f"{name} {units}"
        return name_with_units

    def _create_ggd_name(self, array):
        """Generates a name for the GGD array. The name of the metadata
        of each parent node is stored as well as the identifier, name or labels of the
        node, which are added in brackets, if applicable.

        Args:
            array: The GGD array

        Returns:
            Name of the GGD scalar or vector
        """
        node_id = id(array)
        if node_id in self._cache:
            return self._cache[node_id]
        name = create_name_recursive(array)

        self._cache[node_id] = name
        return name

    def _add_scalar_array_to_vtk_field_data(
        self, array: np.ndarray, name: str, ugrid: vtkUnstructuredGrid
    ) -> None:
        """Add a named array as scalars to a vtkUnstructuredGrid instance

        Args:
            array: the numpy array.
            name: the name string.
            ugrid: an instance of vtkUnstructuredGrid
        """

        logger.debug(f"           {name}...")
        point_data: vtkPointData = ugrid.GetPointData()
        num_points = ugrid.GetNumberOfPoints()
        cell_data: vtkCellData = ugrid.GetCellData()
        num_cells = ugrid.GetNumberOfCells()

        # Split complex arrays into real and imaginary parts
        if array.metadata.data_type is IDSDataType.CPX:
            array_real = np.real(array)
            array_imag = np.imag(array)
            vtk_arr_real = dsa.numpyTovtkDataArray(array_real, name=f"{name}_real")
            vtk_arr_imag = dsa.numpyTovtkDataArray(array_imag, name=f"{name}_imag")
        else:
            vtk_arr = dsa.numpyTovtkDataArray(array, name)

        # To see interpolation of point data on cells, just point data is necessary.
        if len(array) == num_points:
            if array.metadata.data_type is IDSDataType.CPX:
                point_data.AddArray(vtk_arr_real)
                point_data.AddArray(vtk_arr_imag)
            else:
                point_data.AddArray(vtk_arr)
        if len(array) == num_cells:
            if array.metadata.data_type is IDSDataType.CPX:
                cell_data.AddArray(vtk_arr_real)
                cell_data.AddArray(vtk_arr_imag)
            else:
                cell_data.AddArray(vtk_arr)

    def _add_aos_scalar_array_to_vtk_field_data(
        self, aos_scalar_node, subset_idx: int, name: str, ugrid: vtkUnstructuredGrid
    ) -> None:
        """Add the array under the aos_scalar_node to the unstructured grid.

        Args:
            aos_scalar_node: A node with scalar values for each grid subset.
            subset_idx: an index into aos_scalar_node
            name: this becomes the array name in VTK
            ugrid: an unstructured grid instance
        """
        logger.debug(f"           {name}...")
        if subset_idx >= len(aos_scalar_node):
            return

        # For wall IDS nodes, edges, cells, volumes in one partition.
        if subset_idx == -1:
            for i in range(4):
                try:
                    if hasattr(aos_scalar_node[i], "values") and len(
                        aos_scalar_node[i].values
                    ):
                        self._add_scalar_array_to_vtk_field_data(
                            aos_scalar_node[i].values, name, ugrid
                        )
                except IndexError:
                    logger.warning(
                        f"           no index {i} for subset {subset_idx}..."
                    )
                except AttributeError:
                    logger.warning(
                        f"           no index {i} for subset {subset_idx}..."
                    )
        else:
            if hasattr(aos_scalar_node[subset_idx], "values") and len(
                aos_scalar_node[subset_idx].values
            ):
                self._add_scalar_array_to_vtk_field_data(
                    aos_scalar_node[subset_idx].values, name, ugrid
                )

    def _add_aos_vector_array_to_vtk_field_data(
        self, aos_vector_node, subset_idx: int, name: str, ugrid: vtkUnstructuredGrid
    ) -> None:
        """Add the array under the aos_vector_node to the unstructured grid.

        Args:
            aos_vector_node: A node with component vectors for each grid subset.
            subset_idx: an index into aos_vector_node
            name: this becomes the array name in VTK
            ugrid: an unstructured grid instance
        """
        logger.debug(f"           {name}...")
        if subset_idx >= len(aos_vector_node):
            return

        point_data: vtkPointData = ugrid.GetPointData()
        num_points = ugrid.GetNumberOfPoints()
        cell_data: vtkCellData = ugrid.GetCellData()
        num_cells = ugrid.GetNumberOfCells()

        # Only add the components that have data:
        components = dict()  # name and values
        for component_name in [
            "radial",
            "diamagnetic",
            "parallel",
            "poloidal",
            "toroidal",
            "r",
            "z",
        ]:
            try:
                values = getattr(aos_vector_node[subset_idx], component_name)
            except (IndexError, AttributeError):
                continue
            if len(values):
                components[component_name] = values

        vtk_arr = vtkDoubleArray()
        vtk_arr.SetName(name)
        vtk_arr.SetNumberOfComponents(len(components))
        num_tuples = 0

        for i, component_name in enumerate(components):
            vtk_arr.SetComponentName(i, component_name.capitalize())
            scalar_arr = dsa.numpyTovtkDataArray(
                components[component_name], name + "-" + component_name.capitalize()
            )
            if num_tuples == 0:
                num_tuples = scalar_arr.GetNumberOfTuples()
                vtk_arr.SetNumberOfTuples(num_tuples)

            vtk_arr.CopyComponent(i, scalar_arr, 0)

        if num_tuples == num_points:
            point_data.AddArray(vtk_arr)
        if num_tuples == num_cells:
            cell_data.AddArray(vtk_arr)



import imas
import numpy as np
import pytest
from imas import DBEntry
from vtkmodules.vtkCommonDataModel import vtkPolyData

from imas_paraview.plugins.position import PositionReader
from imas_paraview.util import pol_to_cart


@pytest.mark.skip(reason="no IMAS-Core available")
def test_load_position_magnetics():
    """Test if positions of magnetics barometry IDS are saved into vtkPolyData."""
    reader = PositionReader()
    entry = DBEntry(
        (
            "imas:hdf5?path=/work/imas/shared/imasdb/ITER_MACHINE_DESCRIPTION/"
            "/3/150100/5/"
        ),
        "r",
    )
    ids = entry.get("magnetics", lazy=True, autoconvert=False)

    reader._ids = ids
    reader.setup_ids()

    r1 = ids.b_field_pol_probe[0].position.r
    phi1 = ids.b_field_pol_probe[0].position.phi
    z1 = ids.b_field_pol_probe[0].position.z
    name1 = f"{ids.b_field_pol_probe[0].name} / {ids.b_field_pol_probe[0].identifier}"
    point1 = (*pol_to_cart(r1, phi1), z1)
    output = vtkPolyData()
    reader._selected = [name1]
    reader._load_position(output)
    assert output.GetNumberOfPoints() == 1
    assert np.all(np.isclose(point1, output.GetPoint(0)))


@pytest.mark.skip(reason="no IMAS-Core available")
def test_load_position_barometry():
    """Test if positions of gauges of barometry IDS are saved into vtkPolyData."""
    ids = imas.IDSFactory().barometry()
    ids.gauge.resize(2)

    r1 = 1.1
    phi1 = 2.2
    z1 = 3.3
    name1 = "gauge1"
    ids.gauge[0].name = name1
    ids.gauge[0].position.r = r1
    ids.gauge[0].position.phi = phi1
    ids.gauge[0].position.z = z1
    point1 = (*pol_to_cart(r1, phi1), z1)

    r2 = 4.4
    phi2 = 5.5
    z2 = 6.6
    name2 = "gauge2"
    ids.gauge[1].name = name2
    ids.gauge[1].position.r = r2
    ids.gauge[1].position.phi = phi2
    ids.gauge[1].position.z = z2
    point2 = (*pol_to_cart(r2, phi2), z2)

    reader = PositionReader()
    reader._ids = ids
    reader.setup_ids()

    output = vtkPolyData()
    reader._selected = [name1]
    reader._load_position(output)
    assert output.GetNumberOfPoints() == 1
    assert np.all(np.isclose(point1, output.GetPoint(0)))

    output = vtkPolyData()
    reader._selected = [name1, name2]
    reader._load_position(output)
    assert output.GetNumberOfPoints() == 2
    assert np.all(np.isclose(point1, output.GetPoint(0)))
    assert np.all(np.isclose(point2, output.GetPoint(1)))


def test_load_position_langmuir_probes():
    """Test if positions of embedded langmuir probes are saved into vtkPolyData."""
    ids = imas.IDSFactory().langmuir_probes()
    ids.embedded.resize(2)

    r1 = 1.1
    phi1 = 2.2
    z1 = 3.3
    name1 = "probe1"
    ids.embedded[0].name = name1
    ids.embedded[0].position.r = r1
    ids.embedded[0].position.phi = phi1
    ids.embedded[0].position.z = z1
    point1 = (*pol_to_cart(r1, phi1), z1)

    r2 = 4.4
    phi2 = 5.5
    z2 = 6.6
    name2 = "probe2"
    ids.embedded[1].name = name2
    ids.embedded[1].position.r = r2
    ids.embedded[1].position.phi = phi2
    ids.embedded[1].position.z = z2
    point2 = (*pol_to_cart(r2, phi2), z2)

    reader = PositionReader()
    reader._ids = ids
    reader.setup_ids()

    output = vtkPolyData()
    reader._selected = [name1]
    reader._load_position(output)
    assert output.GetNumberOfPoints() == 1
    assert np.all(np.isclose(point1, output.GetPoint(0)))

    output = vtkPolyData()
    reader._selected = [name1, name2]
    reader._load_position(output)
    assert output.GetNumberOfPoints() == 2
    assert np.all(np.isclose(point1, output.GetPoint(0)))
    assert np.all(np.isclose(point2, output.GetPoint(1)))



import imas
import numpy as np
import pytest
from click import UsageError
from click.testing import CliRunner

from imas_paraview.cli import (
    cli,
    convert_ggd_to_vtk,
    parse_index,
    parse_time,
    parse_uri,
)


@pytest.mark.skip(reason="no IMAS-Core available")
def test_version():
    runner = CliRunner()

    result = runner.invoke(cli, ["--version"])
    assert result.exit_code == 0
    assert "IMAS-ParaView version:" in result.output
    assert "IMAS-Python version:" in result.output
    assert "Default data dictionary version:" in result.output
    assert "Available data dictionary versions:" in result.output
    assert "Access Layer core version:" in result.output


def test_ggd2vtk(tmp_path, dummy_ids):
    uri = f"{tmp_path}/testdb.nc"
    with imas.DBEntry(uri, "w") as dbentry:
        dbentry.put(dummy_ids)

    runner = CliRunner()
    file_name = "test"
    output_path = tmp_path / file_name
    ids_name = dummy_ids.metadata.name
    output_str = str(output_path)
    occurrence = "0"
    uri_in = f"{uri}#{ids_name}:{occurrence}"
    args = [uri_in, output_str]
    result = runner.invoke(convert_ggd_to_vtk, args)
    assert result.exit_code == 0

    # The vtkXMLPartitionedDataSetCollectionWriter will output the following:
    # .
    # ├── test
    # │   ├── test_0_0.vtu
    # │   ├── test_1_0.vtu
    # │   ├── test_2_0.vtu
    # |   ├── ...
    # └── test.vtpc

    # Check if vtpc file and the directory containing vtu files exists
    assert output_path.is_dir()
    output_dir = output_path / f"{ids_name}_0"
    assert output_dir.with_suffix(".vtpc").exists()

    if ids_name == "wall":
        vtu_file = output_dir / f"{ids_name}_0_0_0.vtu"
        assert vtu_file.exists()
    else:
        for i in range(3):
            vtu_file = output_dir / f"{ids_name}_0_{i}_0.vtu"
            assert vtu_file.exists()


def test_parse_uri():
    uri_path, ids, occurrence = parse_uri("imas:hdf5?path=testdb#edge_profiles")
    assert uri_path == "imas:hdf5?path=testdb"
    assert ids == "edge_profiles"
    assert occurrence == 0
    uri_path, ids, occurrence = parse_uri("imas:hdf5?path=testdb2#edge_sources:1")
    assert uri_path == "imas:hdf5?path=testdb2"
    assert ids == "edge_sources"
    assert occurrence == 1
    with pytest.raises(UsageError):
        parse_uri("imas:hdf5?path=testdb")


def test_parse_index():
    # Assert correct usage
    valid_cases = [
        ("5", [5]),
        ("3:6", [3, 4, 5, 6]),
        ("2,4,6,7", [2, 4, 6, 7]),
        ("3:3", [3]),
        ("2,4,6,2", [2, 4, 6]),  # Duplicates should be removed
    ]
    for input_str, expected_output in valid_cases:
        assert parse_index(input_str) == expected_output

    # Assert invalid usage (raises UsageError)
    invalid_cases = [
        "",
        ",",
        ":",
        "a:b",
        "a,b,c",
        "3e",
        "5:3",
        "3::5",
        ":5",
        "5:",
        "-1",
        "-3:-1",
        "-5,2,-1",
        "1.2",
        "3:5.0",
        "2,3.0,4",
    ]
    for input_str in invalid_cases:
        with pytest.raises(UsageError):
            parse_index(input_str)


def test_parse_time():
    ids_times = np.array([1.1, 3.3, 6.6, 9.9, 10, 15])

    # Assert correct usage
    valid_cases = [
        ("3.3", [1]),
        ("3.3,6.6,10", [1, 2, 4]),
        ("3:11", [1, 2, 3, 4]),
        ("6.5", [1]),
        ("7,7.2,8.1,9.5", [2]),
        ("5, 14.19", [1, 4]),
        ("14", [4]),
        ("0:14", [0, 1, 2, 3, 4]),
        ("3:9.9", [1, 2, 3]),
        ("9.9:15", [3, 4, 5]),
        ("15", [5]),
        ("0", []),
    ]
    for input_str, expected_output in valid_cases:
        assert parse_time(ids_times, input_str) == expected_output

    # Assert invalid usage (raises UsageError)
    invalid_cases = [
        "",
        ",",
        ":",
        "a:b",
        "a,b,c",
        "3e",
        "5:3",
        "3::5",
        ":5",
        "5:",
    ]
    for input_str in invalid_cases:
        with pytest.raises(UsageError):
            parse_time(ids_times, input_str)



import logging
import os
import sys
from pathlib import Path

import imas
import pytest

from imas_paraview.plugins.vtkggdreader import SUPPORTED_IDS_NAMES
from imas_paraview.tests.fill_ggd import fill_ids


logger = logging.getLogger("imas_paraview")


def set_environment():
    """Set environment variables for PV_PLUGIN_PATH and PYTHONPATH."""
    current_path = Path(__file__).resolve().parent

    # Update PV_PLUGIN_PATH
    pv_plugin_path = os.environ.get("PV_PLUGIN_PATH", "")
    os.environ["PV_PLUGIN_PATH"] = f"{current_path.parent}/plugins:{pv_plugin_path}"

    # Update PYTHONPATH
    os.environ["PYTHONPATH"] = ":".join(sys.path)

    print("Setting the following environment variables:")
    print(f"PV_PLUGIN_PATH={os.environ['PV_PLUGIN_PATH']}")
    print(f"PYTHONPATH={os.environ['PYTHONPATH']}")


def pytest_sessionstart(session):
    """Set up the environment variables and generate test data to
    perform the integration tests on."""

    print("Setting up test environment...")
    set_environment()

    # Generate test data
    DD_VERSION = "3.42.0"
    ids = imas.IDSFactory(version=DD_VERSION).new("edge_profiles")
    fill_ids(ids, time_steps=10, grid_size=5)

    # Write test file as netCDF
    with imas.DBEntry("netcdf_testdb.nc", "w", dd_version=DD_VERSION) as dbentry:
        dbentry.put(ids)

    if not imas.backends.imas_core.imas_interface.has_imas:
        logger.warning(
            "IMAS-Core is not available, some integration tests are skipped."
        )
        return

    # Write test file as MDSPlus
    with imas.DBEntry(
        "imas:mdsplus?path=mdsplus_testdb", "w", dd_version=DD_VERSION
    ) as dbentry:
        dbentry.put(ids)

    # Write test file as HDF5
    with imas.DBEntry(
        "imas:hdf5?path=hdf5_testdb", "w", dd_version=DD_VERSION
    ) as dbentry:
        dbentry.put(ids)

    print("Test environment setup complete.")


@pytest.fixture(params=SUPPORTED_IDS_NAMES)
def ids_name(request):
    """Provides IDS names for testing."""

    return request.param


@pytest.fixture
def dummy_ids(ids_name):
    """Creates a dummy IDS object with a dummy grid and random GGD values for
    testing purposes."""

    ids = imas.IDSFactory().new(ids_name)
    fill_ids(ids)
    ids.validate()
    return ids


@pytest.fixture
def dummy_ids_five_steps(ids_name):
    """Creates a dummy IDS object containing five time steps with a dummy grid and
    random GGD values for testing purposes."""

    ids = imas.IDSFactory().new(ids_name)
    fill_ids(ids, time_steps=5)
    ids.validate()
    return ids
